import ast
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from panther_analysis_tool.schemas import (
    GLOBAL_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    extract_keys_schema,
)

EXPERIMENTAL_STATUS = "experimental"
DEPRECATED_STATUS = "deprecated"


@dataclass
class Filter:
    key: str
    values: List[Any] | Any
    inverted: bool = False


# Parses the filters, expects a list of strings
def parse_filter_args(str_filters: Optional[List[str]]) -> Tuple[List[Filter], List[Filter]]:
    parsed_filters: Dict[str, Any] = {}
    parsed_filters_inverted: Dict[str, Any] = {}
    if str_filters is None:
        return [], []
    for filt in str_filters:
        split = filt.split("=")
        if len(split) != 2 or split[0] == "" or split[1] == "":
            raise ValueError(f"Filter {filt} is not in format KEY=VALUE")
        # Check for "!="
        invert_filter = split[0].endswith("!")
        if invert_filter:
            split[0] = split[0][:-1]  # Remove the trailing "!"
        key = split[0]
        valid_keys = (
            extract_keys_schema(GLOBAL_SCHEMA)
            | extract_keys_schema(POLICY_SCHEMA)
            | extract_keys_schema(RULE_SCHEMA)
        )
        if key not in valid_keys:
            raise ValueError(f"Filter key {key} is not a valid filter field")
        if invert_filter:
            parsed_filters_inverted[key] = split[1].split(",")
        else:
            parsed_filters[key] = split[1].split(",")
        # Handle boolean fields
        if key == "Enabled":
            try:
                bool_value = bool(strtobool(split[1]))
            except ValueError:
                raise ValueError(f"Filter key {key} should have either true or false")
            if invert_filter:
                parsed_filters_inverted[key] = [bool_value]
            else:
                parsed_filters[key] = [bool_value]

    filters = [Filter(key=key, values=values) for key, values in parsed_filters.items()]
    filters_inverted = [
        Filter(key=key, values=values, inverted=True)
        for key, values in parsed_filters_inverted.items()
    ]
    return filters, filters_inverted


def get_filters_with_status_filters(
    str_filters: Optional[List[str]],
) -> Tuple[List[Filter], List[Filter]]:
    filters, filters_inverted = parse_filter_args(str_filters)
    filters, filters_inverted = add_status_filters(filters, filters_inverted)
    return filters, filters_inverted


def add_status_filters(
    filters: List[Filter], filters_inverted: List[Filter]
) -> Tuple[List[Filter], List[Filter]]:
    blocked_statuses = [EXPERIMENTAL_STATUS, DEPRECATED_STATUS]

    # strip experimental/deprecated from any positive Status filter the user provided
    for i, filt in enumerate(filters):
        if filt.key == "Status":
            existing_values = filt.values if isinstance(filt.values, list) else [filt.values]
            cleaned = [v for v in existing_values if v.lower() not in blocked_statuses]
            filters[i] = Filter(key="Status", values=cleaned)

    # ensure the inverted filter always includes experimental and deprecated
    for i, filt in enumerate(filters_inverted):
        if filt.key == "Status":
            existing_values = filt.values if isinstance(filt.values, list) else [filt.values]
            merged_values = list(existing_values)
            for val in blocked_statuses:
                if val not in merged_values:
                    merged_values.append(val)
            filters_inverted[i] = Filter(key="Status", values=merged_values, inverted=True)
            return filters, filters_inverted

    filters_inverted.append(Filter(key="Status", values=list(blocked_statuses), inverted=True))
    return filters, filters_inverted


def strtobool(val: str) -> int:
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))


def collect_top_level_imports(py: bytes) -> set[str]:
    """
    Collects all imports from a Python file by parsing the file is an AST
    and extracting the top level imports.

    Example:
    ```python
        from top import foo
        import bar
        import baz.qux
        from baz.qux import quux
        from scoob.qux import quux as quuux
    ```

    Outputs:
    ```python
        {"top", "bar", "baz", "scoob"}
    ```

    Args:
        py: The Python file to parse.

    Returns:
        A set of top level imports.
    """
    try:
        tree = ast.parse(py.decode("utf-8"))
    except (UnicodeDecodeError, SyntaxError):
        return set()

    imports: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            imports.update(alias.name.split(".")[0] for alias in node.names)
        elif isinstance(node, ast.ImportFrom) and node.module:
            imports.add(node.module.split(".")[0])

    return imports


def search_terms_to_filters(search_terms: list[str]) -> list[Filter]:
    result: list[Filter] = []

    for search_term in search_terms:
        if search_term.strip() == "":
            continue

        try:
            filters, filters_inverted = parse_filter_args([search_term])
            result.extend(filters)
            result.extend(filters_inverted)
        except ValueError:
            result.append(Filter(key="", values=[search_term]))

    return result



# AUTOGENERATED DO NOT EDIT
# Autogenerated using the pat_regex_generator dev tool

from schema import Regex

LOG_TYPE_REGEX = Regex(
    r"^("
    r"AWS\.ALB|"
    r"AWS\.AuroraMySQLAudit|"
    r"AWS\.BedrockModelInvocation|"
    r"AWS\.CloudFrontAccess|"
    r"AWS\.CloudTrail|"
    r"AWS\.CloudTrailDigest|"
    r"AWS\.CloudTrailInsight|"
    r"AWS\.CloudWatchEvents|"
    r"AWS\.Config|"
    r"AWS\.GuardDuty|"
    r"AWS\.NLB|"
    r"AWS\.S3ServerAccess|"
    r"AWS\.SecurityFindingFormat|"
    r"AWS\.TransitGatewayFlow|"
    r"AWS\.VPCDns|"
    r"AWS\.VPCFlow|"
    r"AWS\.WAFWebACL|"
    r"AlphaSOC\.Alert|"
    r"Amazon\.EKS\.Audit|"
    r"Amazon\.EKS\.Authenticator|"
    r"Anomali\.Indicator|"
    r"Apache\.AccessCombined|"
    r"Apache\.AccessCommon|"
    r"AppOmni\.Alerts|"
    r"AppOmni\.Events|"
    r"AppOmni\.Policy|"
    r"Asana\.Audit|"
    r"Atlassian\.Audit|"
    r"Auth0\.Events|"
    r"Axonius\.Activity|"
    r"Azure\.Audit|"
    r"Azure\.DefenderAlerts|"
    r"Azure\.MonitorActivity|"
    r"Bitwarden\.Events|"
    r"Box\.Event|"
    r"CarbonBlack\.AlertV2|"
    r"CarbonBlack\.Audit|"
    r"CarbonBlack\.EndpointEvent|"
    r"CarbonBlack\.WatchlistHit|"
    r"CiscoUmbrella\.AdminAudit|"
    r"CiscoUmbrella\.CloudFirewall|"
    r"CiscoUmbrella\.DNS|"
    r"CiscoUmbrella\.IP|"
    r"CiscoUmbrella\.Proxy|"
    r"Cloudflare\.Audit|"
    r"Cloudflare\.Firewall|"
    r"Cloudflare\.HttpRequest|"
    r"Cloudflare\.Spectrum|"
    r"Cloudflare\.ZeroTrust\.RData|"
    r"Crowdstrike\.AIDMaster|"
    r"Crowdstrike\.ActivityAudit|"
    r"Crowdstrike\.AppInfo|"
    r"Crowdstrike\.CriticalFile|"
    r"Crowdstrike\.DNSRequest|"
    r"Crowdstrike\.DetectionSummary|"
    r"Crowdstrike\.EventStreams|"
    r"Crowdstrike\.FDREvent|"
    r"Crowdstrike\.GroupIdentity|"
    r"Crowdstrike\.ManagedAssets|"
    r"Crowdstrike\.NetworkConnect|"
    r"Crowdstrike\.NetworkListen|"
    r"Crowdstrike\.NotManagedAssets|"
    r"Crowdstrike\.ProcessRollup2|"
    r"Crowdstrike\.ProcessRollup2Stats|"
    r"Crowdstrike\.SyntheticProcessRollup2|"
    r"Crowdstrike\.Unknown|"
    r"Crowdstrike\.UserIdentity|"
    r"Crowdstrike\.UserInfo|"
    r"Crowdstrike\.UserLogonLogoff|"
    r"Docker\.Events|"
    r"Docusign\.Connect|"
    r"Dropbox\.TeamEvent|"
    r"Duo\.Administrator|"
    r"Duo\.Authentication|"
    r"Duo\.OfflineEnrollment|"
    r"Duo\.Telephony|"
    r"Envoy\.Access|"
    r"Fastly\.Access|"
    r"Fluentd\.Syslog3164|"
    r"Fluentd\.Syslog5424|"
    r"GCP\.AccessTransparency|"
    r"GCP\.AuditLog|"
    r"GCP\.DNS|"
    r"GCP\.Firewall|"
    r"GCP\.HTTPLoadBalancer|"
    r"GSuite\.ActivityEvent|"
    r"GSuite\.DirectoryUsers|"
    r"GSuite\.Reports|"
    r"GitHub\.Audit|"
    r"GitHub\.Webhook|"
    r"GitLab\.API|"
    r"GitLab\.Audit|"
    r"GitLab\.Exceptions|"
    r"GitLab\.Git|"
    r"GitLab\.Integrations|"
    r"GitLab\.Production|"
    r"Gravitational\.TeleportAudit|"
    r"GreyNoise\.API\.Noise|"
    r"GreyNoise\.API\.V3|"
    r"GreyNoise\.Noise|"
    r"GreyNoise\.RIOT|"
    r"Heroku\.Runtime|"
    r"IPInfo\.ASNCIDR|"
    r"IPInfo\.ASNRanges|"
    r"IPInfo\.LocationCIDR|"
    r"IPInfo\.LocationRanges|"
    r"IPInfo\.PrivacyCIDR|"
    r"IPInfo\.PrivacyRanges|"
    r"Iru\.Audit|"
    r"Jamfpro\.ComplianceReporter|"
    r"Jamfpro\.Login|"
    r"Juniper\.Access|"
    r"Juniper\.Audit|"
    r"Juniper\.Firewall|"
    r"Juniper\.MWS|"
    r"Juniper\.Postgres|"
    r"Juniper\.Security|"
    r"Kandji\.Endpoint|"
    r"Lacework\.AgentManagement|"
    r"Lacework\.AlertDetails|"
    r"Lacework\.AllFiles|"
    r"Lacework\.Applications|"
    r"Lacework\.ChangeFiles|"
    r"Lacework\.CloudCompliance|"
    r"Lacework\.CloudConfiguration|"
    r"Lacework\.Cmdline|"
    r"Lacework\.Connections|"
    r"Lacework\.ContainerSummary|"
    r"Lacework\.ContainerVulnDetails|"
    r"Lacework\.DNSQuery|"
    r"Lacework\.Events|"
    r"Lacework\.HostVulnDetails|"
    r"Lacework\.Image|"
    r"Lacework\.Interfaces|"
    r"Lacework\.InternalIPA|"
    r"Lacework\.MachineDetails|"
    r"Lacework\.MachineSummary|"
    r"Lacework\.NewHashes|"
    r"Lacework\.Package|"
    r"Lacework\.PodSummary|"
    r"Lacework\.ProcessSummary|"
    r"Lacework\.UserDetails|"
    r"Lacework\.UserLogin|"
    r"Linux\.Auditd|"
    r"MISP\.WarningListsCIDR|"
    r"Material\.NewCaseCreated|"
    r"Microsoft365\.Audit\.AzureActiveDirectory|"
    r"Microsoft365\.Audit\.Exchange|"
    r"Microsoft365\.Audit\.General|"
    r"Microsoft365\.Audit\.SharePoint|"
    r"Microsoft365\.DLP\.All|"
    r"MicrosoftDefenderXDR\.AdvancedHunting|"
    r"MicrosoftGraph\.SecurityAlert|"
    r"MicrosoftIntune\.AuditLogs|"
    r"MicrosoftIntune\.DeviceComplianceOrg|"
    r"MicrosoftIntune\.Devices|"
    r"MicrosoftIntune\.OperationalLogs|"
    r"MicrosoftIntune\.Windows365AuditLogs|"
    r"MongoDB\.OrganizationEvent|"
    r"MongoDB\.ProjectEvent|"
    r"Netskope\.Audit|"
    r"Nginx\.Access|"
    r"Nginx\.Error|"
    r"Notion\.AuditLogs|"
    r"OCSF\.AccountChange|"
    r"OCSF\.ApiActivity|"
    r"OCSF\.ApplicationLifecycle|"
    r"OCSF\.Authentication|"
    r"OCSF\.AuthorizeSession|"
    r"OCSF\.BaseEvent|"
    r"OCSF\.ComplianceFinding|"
    r"OCSF\.ConfigState|"
    r"OCSF\.DatastoreActivity|"
    r"OCSF\.DetectionFinding|"
    r"OCSF\.DeviceConfigStateChange|"
    r"OCSF\.DhcpActivity|"
    r"OCSF\.DnsActivity|"
    r"OCSF\.EmailActivity|"
    r"OCSF\.EmailFileActivity|"
    r"OCSF\.EmailUrlActivity|"
    r"OCSF\.EntityManagement|"
    r"OCSF\.FileActivity|"
    r"OCSF\.FileHosting|"
    r"OCSF\.FtpActivity|"
    r"OCSF\.GroupManagement|"
    r"OCSF\.HttpActivity|"
    r"OCSF\.IncidentFinding|"
    r"OCSF\.InventoryInfo|"
    r"OCSF\.KernelActivity|"
    r"OCSF\.KernelExtension|"
    r"OCSF\.MemoryActivity|"
    r"OCSF\.ModuleActivity|"
    r"OCSF\.NetworkActivity|"
    r"OCSF\.NetworkFileActivity|"
    r"OCSF\.NtpActivity|"
    r"OCSF\.PatchState|"
    r"OCSF\.ProcessActivity|"
    r"OCSF\.RdpActivity|"
    r"OCSF\.ScanActivity|"
    r"OCSF\.ScheduledJobActivity|"
    r"OCSF\.SecurityFinding|"
    r"OCSF\.SmbActivity|"
    r"OCSF\.SshActivity|"
    r"OCSF\.UserAccess|"
    r"OCSF\.UserInventory|"
    r"OCSF\.VulnerabilityFinding|"
    r"OCSF\.WebResourceAccessActivity|"
    r"OCSF\.WebResourcesActivity|"
    r"OCSF\.WinPrefetchInfo|"
    r"OCSF\.WinRegistryKeyActivity|"
    r"OCSF\.WinRegistryKeyInfo|"
    r"OCSF\.WinRegistryValueActivity|"
    r"OCSF\.WinRegistryValueInfo|"
    r"OCSF\.WinResourceActivity|"
    r"OCSF\.WinprefetchInfo|"
    r"OCSF\.WinregistryKeyActivity|"
    r"OCSF\.WinregistryKeyInfo|"
    r"OCSF\.WinregistryValueActivity|"
    r"OCSF\.WinregistryValueInfo|"
    r"OCSF\.WinresourceActivity|"
    r"OSSEC\.EventInfo|"
    r"OTX\.Pulses|"
    r"Okta\.Devices|"
    r"Okta\.SystemLog|"
    r"Okta\.Users|"
    r"OneLogin\.Events|"
    r"OnePassword\.AuditEvent|"
    r"OnePassword\.ItemUsage|"
    r"OnePassword\.SignInAttempt|"
    r"OpenAI\.Audit|"
    r"Orca\.Alert|"
    r"Orca\.AlertEvent|"
    r"Osquery\.Batch|"
    r"Osquery\.Differential|"
    r"Osquery\.Snapshot|"
    r"Osquery\.Status|"
    r"Panther\.Audit|"
    r"Proofpoint\.Event|"
    r"PushSecurity\.Activity|"
    r"PushSecurity\.Controls|"
    r"PushSecurity\.Entities|"
    r"Rapid7\.AuditLog|"
    r"Salesforce\.Login|"
    r"Salesforce\.LoginAs|"
    r"Salesforce\.Logout|"
    r"Salesforce\.RealtimeEvent|"
    r"Salesforce\.URI|"
    r"SentinelOne\.Activity|"
    r"SentinelOne\.DeepVisibility|"
    r"SentinelOne\.DeepVisibilityV2|"
    r"Slack\.AccessLogs|"
    r"Slack\.AuditLogs|"
    r"Slack\.IntegrationLogs|"
    r"Snowflake\.AccessHistory|"
    r"Snowflake\.DataTransferHistory|"
    r"Snowflake\.GrantsToRoles|"
    r"Snowflake\.GrantsToUsers|"
    r"Snowflake\.LoginHistory|"
    r"Snowflake\.NetworkPolicies|"
    r"Snowflake\.QueryHistory|"
    r"Snowflake\.Roles|"
    r"Snowflake\.Sessions|"
    r"Snowflake\.Stages|"
    r"Snowflake\.Users|"
    r"Snyk\.GroupAudit|"
    r"Snyk\.OrgAudit|"
    r"Sophos\.Central|"
    r"Sublime\.Audit|"
    r"Sublime\.MDM|"
    r"Sublime\.MessageEvent|"
    r"Suricata\.Alert|"
    r"Suricata\.Anomaly|"
    r"Suricata\.DHCP|"
    r"Suricata\.DNS|"
    r"Suricata\.FileInfo|"
    r"Suricata\.Flow|"
    r"Suricata\.HTTP|"
    r"Suricata\.SSH|"
    r"Suricata\.TLS|"
    r"Sysdig\.Audit|"
    r"Syslog\.RFC3164|"
    r"Syslog\.RFC5424|"
    r"Tailscale\.Audit|"
    r"Tailscale\.Network|"
    r"Tenable\.Vulnerability|"
    r"ThinkstCanary\.Alert|"
    r"Tines\.Audit|"
    r"Tor\.ExitNode|"
    r"Tracebit\.Alert|"
    r"Tracebit\.HealthCheck|"
    r"TrailDiscover\.CloudTrail|"
    r"Upwind\.Detections|"
    r"Vercel\.Audit|"
    r"Windows\.EventLogs|"
    r"Wiz\.Audit|"
    r"Wiz\.Detections|"
    r"Wiz\.Issues|"
    r"Wiz\.IssuesWebhook|"
    r"Wiz\.Threats|"
    r"Wiz\.VulnerabilityFinding|"
    r"Workday\.Activity|"
    r"Workday\.SignOnAttempt|"
    r"Zeek\.CaptureLoss|"
    r"Zeek\.Conn|"
    r"Zeek\.DHCP|"
    r"Zeek\.DNS|"
    r"Zeek\.DPD|"
    r"Zeek\.Files|"
    r"Zeek\.HTTP|"
    r"Zeek\.KnownHosts|"
    r"Zeek\.KnownServices|"
    r"Zeek\.NTP|"
    r"Zeek\.Notice|"
    r"Zeek\.OCSP|"
    r"Zeek\.Reporter|"
    r"Zeek\.SIP|"
    r"Zeek\.Software|"
    r"Zeek\.Ssh|"
    r"Zeek\.Ssl|"
    r"Zeek\.Stats|"
    r"Zeek\.Telemetry|"
    r"Zeek\.Tunnel|"
    r"Zeek\.Weird|"
    r"Zeek\.X509|"
    r"Zendesk\.Audit|"
    r"Zoom\.Activity|"
    r"Zoom\.Operation|"
    r"Zscaler\.ZIA\.AdminAuditLog|"
    r"Zscaler\.ZIA\.DNSLog|"
    r"Zscaler\.ZIA\.FWLog|"
    r"Zscaler\.ZIA\.WebLog|"
    r"Zscaler\.ZPA\.AppConnectorMetrics|"
    r"Zscaler\.ZPA\.AppConnectorStatus|"
    r"Zscaler\.ZPA\.AuditLog|"
    r"Zscaler\.ZPA\.UserActivity|"
    r"Zscaler\.ZPA\.UserStatus|"
    r"Custom\.([A-Z][A-Za-z0-9]*)(\.[A-Z][A-Za-z0-9]*){0,5}"
    r")$"
)

RESOURCE_TYPE_REGEX = Regex(
    r"^("
    r"AWS\.ACM\.Certificate|"
    r"AWS\.CloudFormation\.Stack|"
    r"AWS\.CloudTrail|"
    r"AWS\.CloudTrail\.Meta|"
    r"AWS\.CloudWatch\.LogGroup|"
    r"AWS\.Config\.Recorder|"
    r"AWS\.Config\.Recorder\.Meta|"
    r"AWS\.DynamoDB\.Table|"
    r"AWS\.EC2\.AMI|"
    r"AWS\.EC2\.Instance|"
    r"AWS\.EC2\.NetworkACL|"
    r"AWS\.EC2\.SecurityGroup|"
    r"AWS\.EC2\.VPC|"
    r"AWS\.EC2\.Volume|"
    r"AWS\.ECS\.Cluster|"
    r"AWS\.EKS\.Cluster|"
    r"AWS\.ELBV2\.ApplicationLoadBalancer|"
    r"AWS\.GuardDuty\.Detector|"
    r"AWS\.GuardDuty\.Detector\.Meta|"
    r"AWS\.IAM\.Group|"
    r"AWS\.IAM\.Policy|"
    r"AWS\.IAM\.Role|"
    r"AWS\.IAM\.RootUser|"
    r"AWS\.IAM\.User|"
    r"AWS\.KMS\.Key|"
    r"AWS\.Lambda\.Function|"
    r"AWS\.PasswordPolicy|"
    r"AWS\.RDS\.Instance|"
    r"AWS\.Redshift\.Cluster|"
    r"AWS\.Route53\.HostedZone|"
    r"AWS\.Route53Domains|"
    r"AWS\.S3\.Bucket|"
    r"AWS\.WAF\.Regional\.WebACL|"
    r"AWS\.WAF\.WebACL|"
    r")$"
)



import logging
from typing import Tuple

from panther_analysis_tool.backend.client import Client as BackendClient


def run(backend: BackendClient, api_host: str) -> Tuple[int, str]:
    logging.info("checking connection to %s...", api_host)
    result = backend.check()

    if not result.success:
        logging.info("connection failed")
        return 1, result.message

    logging.info("connection successful: %s", result.message)
    return 0, ""



"""
A module for migrating analysis items from v1 of Panther Analysis usage, where a user
forks the Panther Analysis repository, to v2 where a user uses the Panther Analysis tool to
manage their analysis items without a fork, taking over the merge process.
"""

import dataclasses
import io
import json
import logging
import pathlib
from typing import Tuple

from rich.progress import BarColumn, Progress, TextColumn, track
from ruamel.yaml import scanner  # from tqdm

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.constants import (
    CACHED_MIGRATION_STATUS_FILE_PATH,
    MIGRATION_STATUS_FILE_PATH,
    AnalysisTypes,
    AutoAcceptOption,
)
from panther_analysis_tool.core import (
    analysis_cache,
    file_editor,
    git_helpers,
    install_item,
    merge_item,
    root,
    versions_file,
    yaml,
)


@dataclasses.dataclass
class MigrateArgs:
    analysis_id: str | None = None
    editor: str | None = None
    auto_accept: AutoAcceptOption | None = None
    write_merge_conflicts: bool = False


@dataclasses.dataclass
class MigrationItem:
    """
    An item that was migrated or should be migrated.

    Attributes:
        analysis_id: The ID of the analysis item
        pretty_analysis_type: The type of the analysis item in a pretty format
        merged_item: The merged item that was migrated
        reason: More information about what happened to the item during migration
    """

    analysis_id: str
    "The ID of the analysis item"

    pretty_analysis_type: str
    "The type of the analysis item in a pretty format"

    merged_item: analysis_utils.AnalysisItem | None = None
    "The merged item that was migrated, or None if there was a merge conflict"

    reason: str | None = None
    "More information about what happened to the item during migration"

    def to_dict(self) -> dict:
        """Returns a JSON string representation of the migration item without the merged item."""
        return {
            "analysis_id": self.analysis_id,
            "pretty_analysis_type": self.pretty_analysis_type,
            "reason": self.reason,
        }

    @staticmethod
    def from_dict(dict_: dict) -> "MigrationItem":
        """Returns a MigrationItem from a JSON dictionary."""
        return MigrationItem(
            analysis_id=dict_.get("analysis_id", ""),
            pretty_analysis_type=dict_.get("pretty_analysis_type", ""),
            reason=dict_.get("reason", None),
        )


@dataclasses.dataclass
class MigrationStatus:
    """
    Status of the user's migration.

    Attributes:
        items_with_conflicts (dict[str, MigrationItem]): Items not migrated due to merge conflicts
        items_migrated (dict[str, MigrationItem]): Items migrated
        items_deleted (dict[str, MigrationItem]): Items deleted during the migration
    """

    items_with_conflicts: dict[str, MigrationItem] = dataclasses.field(default_factory=dict)
    "Items not migrated due to merge conflicts"

    items_migrated: dict[str, MigrationItem] = dataclasses.field(default_factory=dict)
    "Items migrated"

    items_deleted: dict[str, MigrationItem] = dataclasses.field(default_factory=dict)
    "Items deleted"

    items_with_warnings: dict[str, MigrationItem] = dataclasses.field(default_factory=dict)
    "Items with warnings"

    def empty(self) -> bool:
        return (
            len(self.items_with_conflicts) == 0
            and len(self.items_migrated) == 0
            and len(self.items_deleted) == 0
            and len(self.items_with_warnings) == 0
        )

    def has_conflicts(self) -> bool:
        return len(self.items_with_conflicts) > 0

    def has_warnings(self) -> bool:
        return len(self.items_with_warnings) > 0

    def items_with_conflicts_sorted(self) -> list[MigrationItem]:
        return self._sort_items(self.items_with_conflicts)

    def items_deleted_sorted(self) -> list[MigrationItem]:
        return self._sort_items(self.items_deleted)

    def items_migrated_sorted(self) -> list[MigrationItem]:
        return self._sort_items(self.items_migrated)

    def items_with_warnings_sorted(self) -> list[MigrationItem]:
        return self._sort_items(self.items_with_warnings)

    def _sort_items(self, items: dict[str, MigrationItem]) -> list[MigrationItem]:
        return sorted(items.values(), key=lambda x: x.pretty_analysis_type + x.analysis_id)

    def migrated_merged_items(self) -> list[analysis_utils.AnalysisItem]:
        return [
            item.merged_item
            for item in self.items_migrated.values()
            if item.merged_item is not None
        ]

    def to_dict(self) -> dict:
        """Returns a JSON string representation of the migration status."""
        return {
            "items_with_conflicts": {k: v.to_dict() for k, v in self.items_with_conflicts.items()},
            "items_migrated": {k: v.to_dict() for k, v in self.items_migrated.items()},
            "items_deleted": {k: v.to_dict() for k, v in self.items_deleted.items()},
            "items_with_warnings": {k: v.to_dict() for k, v in self.items_with_warnings.items()},
        }

    @staticmethod
    def from_dict(dict_: dict) -> "MigrationStatus":
        """Returns a MigrationStatus from a JSON dictionary."""
        return MigrationStatus(
            items_with_conflicts={
                k: MigrationItem.from_dict(v)
                for k, v in dict_.get("items_with_conflicts", {}).items()
            },
            items_migrated={
                k: MigrationItem.from_dict(v) for k, v in dict_.get("items_migrated", {}).items()
            },
            items_deleted={
                k: MigrationItem.from_dict(v) for k, v in dict_.get("items_deleted", {}).items()
            },
            items_with_warnings={
                k: MigrationItem.from_dict(v)
                for k, v in dict_.get("items_with_warnings", {}).items()
            },
        )

    def write_migration_status(self) -> None:
        if self.empty():
            return

        # write the cached version of the migration status as json
        CACHED_MIGRATION_STATUS_FILE_PATH.parent.mkdir(parents=True, exist_ok=True)
        CACHED_MIGRATION_STATUS_FILE_PATH.write_text(json.dumps(self.to_dict()))

        # write the human readable version of the migration status to the file system as a markdown file
        stream = io.StringIO()
        stream.write("# Migration Results\n\n")

        stream.write("## Migration Summary\n\n")
        stream.write(f"  * {len(self.items_with_conflicts)} merge conflict(s) found.\n")
        stream.write(f"  * {len(self.items_deleted)} analysis item(s) deleted.\n")
        stream.write(f"  * {len(self.items_migrated)} analysis item(s) migrated.\n\n")

        if len(self.items_with_conflicts) > 0:
            stream.write("## Analysis Items with Merge Conflicts\n\n")
            stream.write(
                f"{len(self.items_with_conflicts)} merge conflict(s) found. Run `EDITOR=<editor> pat migrate <id>` to resolve each conflict.\n\n"  # pylint: disable=line-too-long
            )
            for conflict in self.items_with_conflicts_sorted():
                stream.write(f"  * ({conflict.pretty_analysis_type}) {conflict.analysis_id}\n")
            stream.write("\n")

        if len(self.items_deleted) > 0:
            stream.write("## Analysis Items Deleted\n\n")
            stream.write(f"{len(self.items_deleted)} analysis item(s) deleted.\n\n")
            for item in self.items_deleted_sorted():
                stream.write(
                    f"  * ({item.pretty_analysis_type}) {item.analysis_id}{f' - {item.reason}' if item.reason is not None else ''}\n"
                )
            stream.write("\n")

        if len(self.items_with_warnings) > 0:
            stream.write("## Analysis Items with Warnings\n\n")
            stream.write(f"{len(self.items_with_warnings)} analysis item(s) with warnings.\n\n")
            for item in self.items_with_warnings_sorted():
                stream.write(
                    f"  * ({item.pretty_analysis_type}) {item.analysis_id}{f' - {item.reason}' if item.reason is not None else ''}\n"
                )
            stream.write("\n")

        if len(self.items_migrated) > 0:
            stream.write("## Analysis Items Migrated\n\n")
            stream.write(f"{len(self.items_migrated)} analysis item(s) migrated.\n\n")
            for item in self.items_migrated_sorted():
                stream.write(f"  * ({item.pretty_analysis_type}) {item.analysis_id}\n")
            stream.write("\n")

        MIGRATION_STATUS_FILE_PATH.write_text(stream.getvalue(), encoding="utf-8")

    @staticmethod
    def load_migration_status() -> "MigrationStatus":
        """
        Load the migration status from the cached file.
        """
        if not CACHED_MIGRATION_STATUS_FILE_PATH.exists():
            return MigrationStatus()
        dict_status = json.loads(CACHED_MIGRATION_STATUS_FILE_PATH.read_text())
        return MigrationStatus.from_dict(dict_status)


def run(args: MigrateArgs) -> Tuple[int, str]:
    root.chdir_to_project_root()
    analysis_cache.update_with_latest_panther_analysis(show_progress_bar=True)

    try:
        migration_result = migrate(
            args.analysis_id,
            args.editor,
            args.auto_accept,
            args.write_merge_conflicts,
        )
    except file_editor.EditorCommandNotFoundError as err:
        return 1, str(err)

    if args.analysis_id is not None:
        # skip the completion message if the user specified an analysis id
        return 0, ""

    if not migration_result.empty():
        completion_message = "Migration complete! Details can be found in: `migration_status.md`."
        if migration_result.has_conflicts():
            completion_message = (
                "Migration completed! However, some analysis items had merge conflicts. "
                "Merge conflicts need to be resolved manually. \n"
                "  * Details can be found in: `migration_status.md`.\n"
                "  * Run `EDITOR=<editor> pat migrate <id>` to resolve each conflict.\n"
                "  * Run `pat migrate --auto-accept=<panthers|yours>` to automatically accept your changes or Panther's changes for merge conflicts."  # pylint: disable=line-too-long
            )
        print(completion_message)
        print()
        print("Run `git diff` to see any changes made.")
    else:
        print("All analysis items in your repo are already migrated! ")
        print("  * Run `pat update` to update to the latest Panther Analysis items.")
        print("  * Run `pat explore` to explore available analysis items.")
        print("  * Run `pat install` to install and enable analysis items you want to use.")
        print()
        print("Run `pat --help` for more information.")

    return 0, ""


def migrate(  # pylint: disable=too-many-locals
    analysis_id: str | None,
    editor: str | None,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
) -> MigrationStatus:
    result = MigrationStatus.load_migration_status()
    cache = analysis_cache.AnalysisCache()

    ancestor_commit: str | None = None

    with Progress(
        TextColumn("Checking when repo last updated:"),
        BarColumn(),
        transient=True,
    ) as progress:
        task = progress.add_task("checking_for_common_ancestor", total=None)
        try:
            ancestor_commit = git_helpers.get_forked_panther_analysis_common_ancestor()
        except RuntimeError as err:
            logging.debug("Failed to get forked panther analysis common ancestor: %s", err)
        progress.update(task, completed=True)

    deprecated_ids = load_deprecated_ids()

    # load all user analysis specs
    specs: list[analysis_utils.LoadAnalysisSpecsResult] = []
    for spec in track(
        analysis_utils.load_analysis_specs_ex(["."], [], True),
        description="Loading user analysis items:",
        disable=analysis_id is not None,
        transient=True,
    ):
        if spec.analysis_spec is None:
            logging.warning(
                "Invalid YAML spec at %s, not migrating",
                spec.spec_filename,
            )
            continue

        if analysis_id is None and spec.analysis_type() == AnalysisTypes.PACK:
            spec.unlink()
            result.items_deleted[spec.analysis_id()] = MigrationItem(
                analysis_id=spec.analysis_id(),
                pretty_analysis_type=spec.pretty_analysis_type(),
                reason="Packs are managed by Panther and not needed in your repo.",
            )

            continue

        if analysis_id is None and spec.analysis_id() in deprecated_ids and not spec.enabled():
            spec.unlink()
            result.items_with_warnings.pop(spec.analysis_id(), None)
            result.items_deleted[spec.analysis_id()] = MigrationItem(
                analysis_id=spec.analysis_id(),
                pretty_analysis_type=spec.pretty_analysis_type(),
                reason="Item was deprecated by Panther and was disabled in your repo.",
            )
            continue

        if analysis_id is None and spec.analysis_id() in deprecated_ids and spec.enabled():
            result.items_with_warnings[spec.analysis_id()] = MigrationItem(
                analysis_id=spec.analysis_id(),
                pretty_analysis_type=spec.pretty_analysis_type(),
                reason="Item was deprecated by Panther and is enabled in your repo. Please consider deleting it.",
            )
            continue

        specs.append(spec)

    # migrate each user analysis spec
    for user_spec in track(
        specs, description="Migration progress:", disable=analysis_id is not None, transient=True
    ):
        item = get_migration_item(user_spec, analysis_id, cache, result, ancestor_commit)
        if item is None:
            continue

        migrate_item(
            item, analysis_id is not None, editor, result, auto_accept, write_merge_conflicts
        )
        ensure_python_file_exists(item)

    # we need to check if the new migrated python includes any
    # new global helper imports and install those so the new python works
    with Progress(
        TextColumn("Installing dependencies:"),
        BarColumn(),
        transient=True,
        disable=analysis_id is not None,
    ) as progress:
        task = progress.add_task("installing_dependencies", total=None)
        items = result.migrated_merged_items()
        install_item.install_deps(items)
        progress.update(task, completed=True)

    with Progress(
        TextColumn("Writing migration results:"),
        BarColumn(),
        transient=True,
        disable=analysis_id is not None,
    ) as progress:
        task = progress.add_task("writing_migration_results", total=None)
        result.write_migration_status()
        progress.update(task, completed=True)

    return result


def get_migration_item(
    user_spec: analysis_utils.LoadAnalysisSpecsResult,
    analysis_id: str | None,
    cache: analysis_cache.AnalysisCache,
    result: MigrationStatus,
    ancestor_commit: str | None = None,
) -> merge_item.MergeableItem | None:
    if analysis_id is not None and analysis_id != user_spec.analysis_id():
        # user specified an analysis id, only migrate that one
        return None

    if user_spec.analysis_type() == AnalysisTypes.PACK:
        # ignoring packs from migration
        return None

    user_spec_id = user_spec.analysis_id()

    # if the user spec does not have a BaseVersion, it needs to be migrated
    if "BaseVersion" in user_spec.analysis_spec:
        if analysis_id is not None:
            print(f"{user_spec_id} already migrated.")
        return None

    base_item = get_base_item(user_spec, ancestor_commit)

    latest_item: analysis_utils.AnalysisItem | None = None
    latest_version: int | None = None

    # load the latest analysis item from the cache using the user spec's ID
    latest_spec = cache.get_latest_spec(user_spec_id)

    # the latest item is simply the latest spec if it was in the cache
    if latest_spec is not None:
        # New loader per spec so roundtrip state is not shared across loads (see load_analysis_specs_ex).
        yaml_loader = yaml.BlockStyleYAML()
        latest_item = analysis_utils.AnalysisItem(
            yaml_file_contents=yaml_loader.load(latest_spec.spec),
            raw_yaml_file_contents=latest_spec.spec,
            python_file_contents=cache.get_file_for_spec(latest_spec.id or -1, latest_spec.version),
        )
        latest_version = latest_spec.version

    # if the latest spec is empty but the base exists, there is a chance panther changed the ID
    # but it is still at the same path, so we can check to see if the latest item is in the remote upstream
    if latest_item is None and not base_item.empty():
        logging.debug(
            "Latest item not in cache, but base item exists, checking for latest item in remote upstream by path: %s",
            user_spec.spec_filename,
        )
        latest_item = get_latest_item_by_path(pathlib.Path(user_spec.spec_filename))

    # The above "if" block is there to handle this edge case, where the item changed IDs but is still at the same path.
    # Sadly, the base_item will not be empty if the user's repo was cloned from a forked repo, because we have an ancestor
    # commit to use. But base_item will be empty if the user's repo was not cloned from forked PA. In that case, we will just
    # assume we need to check for the special edge case of the one analysis item that changed IDs. The above "if" block will
    # be left still to cover anything unexpected, just in case.
    if (
        latest_item is None
        and base_item.empty()
        and user_spec.analysis_id() == "Standard.GSuite.Reports"
    ):
        logging.debug(
            "Latest item not in cache, base item is empty, and user spec ID is Standard.GSuite.Reports, checking for latest item in remote upstream by path: %s",  # pylint: disable=line-too-long
            user_spec.spec_filename,
        )
        latest_item = get_latest_item_by_path(pathlib.Path(user_spec.spec_filename))

    # was deleted by Panther, delete if the item was unused by the user
    if latest_item is None and not base_item.empty() and not user_spec.enabled():
        user_spec.unlink()
        result.items_deleted[user_spec_id] = MigrationItem(
            analysis_id=user_spec_id,
            pretty_analysis_type=user_spec.pretty_analysis_type(),
            reason="Item was deleted by Panther since your last update and was disabled in your repo.",
        )

    # The above "if" block is there to handle this edge case, where the item is no longer in PA, but was in PA when the user last updated,
    # and is disabled in the user's repo. At that point, we assume the user only has it because PA made them have, but they do not actually
    # use it. So we can just delete it. This covered the edge case for the rule "Box.Item.Shared.Externally", where a global helper it
    # was using was removed from PA, but the rule was never updated because the rule was removed from PA, too.
    # Sadly, if the user's repo was not cloned from forked PA, but just a custom repo, we won't have a base_item because we don't have an
    # ancestor commit. In that case, we will just assume we need to check for the special edge case of the one analysis item we know of that
    # has this issue, "Box.Item.Shared.Externally". The above "if" block will be left still to cover anything unexpected, just in case.
    python_file_contents = user_spec.python_file_contents()
    if (
        latest_item is None
        and not user_spec.enabled()
        and user_spec.analysis_id() == "Box.Item.Shared.Externally"
        and python_file_contents is not None
        and "panther_box_helpers" in python_file_contents.decode("utf-8")
    ):
        user_spec.unlink()
        result.items_deleted[user_spec_id] = MigrationItem(
            analysis_id=user_spec_id,
            pretty_analysis_type=user_spec.pretty_analysis_type(),
            reason="Item was deleted by Panther, was disabled in your repo, "
            "and was using the global helper 'panther_box_helpers' which was significantly changed by Panther. "
            "This likely would have caused test failures so it was deleted for a smoother migration.",
        )

    # this happens with custom analysis items
    # or if the item was removed by Panther
    if latest_item is None:
        return None

    # if the latest spec was not in the cache, we don't have a version number yet,
    # so we need to get it from the versions file
    if latest_version is None:
        versions = versions_file.get_versions()
        latest_version = versions.versions[latest_item.analysis_id()].version

    # load the python file for the user spec from the file system
    user_py: bytes | None = None
    py_path: pathlib.Path | None = None
    if user_spec.analysis_spec.get("Filename") is not None:
        py_path = user_spec.python_file_path()
        user_py = py_path.read_bytes() if py_path is not None else None

    return merge_item.MergeableItem(
        user_item=analysis_utils.AnalysisItem(
            yaml_file_contents=user_spec.analysis_spec,
            raw_yaml_file_contents=user_spec.raw_spec_file_content,
            yaml_file_path=user_spec.spec_filename,
            python_file_contents=user_py,
            python_file_path=str(py_path) if py_path is not None else None,
        ),
        latest_panther_item=latest_item,
        base_panther_item=base_item,
        latest_item_version=latest_version,
    )


def migrate_item(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    item: merge_item.MergeableItem,
    solve_merge: bool,
    editor: str | None,
    migration_result: MigrationStatus,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
) -> None:
    yaml_loader = yaml.BlockStyleYAML()

    has_conflict = merge_item.merge_item(
        mergeable_item=item,
        solve_merge=solve_merge,
        editor=editor,
        auto_accept=auto_accept,
        write_merge_conflicts=write_merge_conflicts,
    )
    item_id = item.user_item.analysis_id()
    if has_conflict:
        migration_result.items_migrated.pop(item_id, None)
        migration_result.items_with_conflicts[item_id] = MigrationItem(
            analysis_id=item_id,
            pretty_analysis_type=item.user_item.pretty_analysis_type(),
        )
    else:
        migration_result.items_with_conflicts.pop(item_id, None)
        migration_result.items_migrated[item_id] = MigrationItem(
            analysis_id=item_id,
            pretty_analysis_type=item.user_item.pretty_analysis_type(),
            merged_item=item.merged_item,
        )

    if not has_conflict or write_merge_conflicts:
        # once we know there are no conflicts, or we wrote out the merge conflicts,
        # we can add the BaseVersion to say migration is complete
        yaml_file_path = pathlib.Path(item.user_item.yaml_file_path or "")
        try:
            user_spec: dict = yaml_loader.load(yaml_file_path)
            user_spec["BaseVersion"] = item.latest_item_version
            yaml_loader.dump(user_spec, yaml_file_path)
        except scanner.ScannerError:
            # the yaml might have failed to parse because we wrote out the merge conflicts,
            # but we still need to add the BaseVersion to say migration is complete so we have
            # to add it directly to the text file
            raw_user_spec = (
                yaml_file_path.read_text(encoding="utf-8").strip()
                + f"\nBaseVersion: {item.latest_item_version}\n"
            )
            yaml_file_path.write_text(raw_user_spec, encoding="utf-8")


def get_base_item(
    user_spec: analysis_utils.LoadAnalysisSpecsResult, ancestor_commit: str | None
) -> analysis_utils.AnalysisItem:
    """
    Get the base item for a user spec from Panther Analysis using the ancestor commit.
    If the ancestor commit is not provided, the base item will be an empty analysis item.

    Args:
        user_spec: The user spec to get the base item for
        ancestor_commit: The commit hash of the ancestor commit
    Returns:
        The base item for the user spec
    """
    base_item = analysis_utils.AnalysisItem({}, raw_yaml_file_contents=b"{}")
    if ancestor_commit is None or ancestor_commit == "":
        return base_item

    yaml_file_path = pathlib.Path(user_spec.spec_filename).relative_to(git_helpers.git_root())
    py_file_path = user_spec.python_file_path()
    if py_file_path is not None:
        py_file_path = pathlib.Path(py_file_path).relative_to(git_helpers.git_root())

    # use panther's file path if it exists
    if versions_file.get_versions().has_item(user_spec.analysis_id()):
        history_item = versions_file.get_versions().get_current_version_history_item(
            user_spec.analysis_id()
        )
        yaml_file_path = pathlib.Path(history_item.yaml_file_path)
        if py_file_path is not None:
            py_file_path = pathlib.Path(history_item.py_file_path or "")

    yaml_loader = yaml.BlockStyleYAML()
    base_yaml = git_helpers.get_file_at_commit(ancestor_commit, yaml_file_path)
    if base_yaml is not None:
        base_item.yaml_file_contents = yaml_loader.load(base_yaml)
        base_item.raw_yaml_file_contents = base_yaml

    if py_file_path is not None:
        base_item.python_file_contents = git_helpers.get_file_at_commit(
            ancestor_commit, py_file_path
        )

    return base_item


def ensure_python_file_exists(item: merge_item.MergeableItem) -> None:
    """
    Ensure the python file exists for the merged item.
    """
    if item.merged_item is None:
        return

    if "Filename" not in item.merged_item.yaml_file_contents:
        return

    if item.merged_item.yaml_file_path is None:
        return

    abs_python_file_path: pathlib.Path = (
        pathlib.Path(item.merged_item.yaml_file_path).parent
        / item.merged_item.yaml_file_contents["Filename"]
    )

    # when the item was updated during migration, the "Filename" value might have changed to
    # something not in the user's repo, so we need to check
    if abs_python_file_path.exists():
        return

    # turn absolute path into relative path
    git_python_file_path = abs_python_file_path.relative_to(git_helpers.git_root())

    # if the python file does not exist, try to grab it from PA
    python_file_contents = git_helpers.get_file_at_commit(
        git_helpers.panther_analysis_remote_upstream_branch(), git_python_file_path
    )

    # this is a best attempt effort so it is okay if it does not exist
    if python_file_contents is None or python_file_contents == b"":
        logging.debug(
            "Python file did not exist, and did not exist in remote upstream: %s",
            git_python_file_path,
        )
        return

    # save the remote version of the python file if it does exist
    abs_python_file_path.parent.mkdir(parents=True, exist_ok=True)
    abs_python_file_path.write_bytes(python_file_contents)

    # update the merged item with the new stuff
    item.merged_item.python_file_contents = python_file_contents
    item.merged_item.python_file_path = str(abs_python_file_path)

    logging.debug(
        "Python file did not exist, created it from remote upstream: %s",
        item.merged_item.python_file_path,
    )


def get_latest_item_by_path(spec_path: pathlib.Path) -> analysis_utils.AnalysisItem | None:
    """
    Get the latest version of an analysis item from the remote upstream using the
    spec path of an analysis item in the user's repo. If the spec path is not found
    in the remote upstream, return None.
    """
    git_spec_file_path = spec_path.relative_to(git_helpers.git_root())
    spec_file_contents = git_helpers.get_file_at_commit(
        git_helpers.panther_analysis_remote_upstream_branch(), git_spec_file_path
    )
    if spec_file_contents is None or spec_file_contents == b"":
        logging.debug(
            "Spec file did not exist, and did not exist in remote upstream: %s",
            git_spec_file_path,
        )
        return None

    yaml_loader = yaml.BlockStyleYAML()
    spec = yaml_loader.load(spec_file_contents)

    item = analysis_utils.AnalysisItem(
        yaml_file_contents=spec,
        raw_yaml_file_contents=spec_file_contents,
        yaml_file_path=str(spec_path),
    )

    if "Filename" in spec:
        py_file_path = git_spec_file_path.parent / spec["Filename"]
        py_file_contents = git_helpers.get_file_at_commit(
            git_helpers.panther_analysis_remote_upstream_branch(), py_file_path
        )
        if py_file_contents is not None and py_file_contents != b"":
            item.python_file_contents = py_file_contents
            item.python_file_path = str(spec_path.parent / spec["Filename"])

    logging.debug("Latest item found by path: %s", git_spec_file_path)

    return item


def load_deprecated_ids() -> set[str]:
    """
    Load deprecated analysis IDs from deprecated.txt file in the project root.
    Each line in the file should contain a single analysis ID.

    Returns:
        A list of deprecated analysis IDs, or an empty list if the file doesn't exist.
    """
    deprecated_file = pathlib.Path("deprecated.txt")
    if not deprecated_file.exists():
        return set()

    deprecated_ids = set()
    try:
        content = deprecated_file.read_text(encoding="utf-8")
        for line in content.splitlines():
            line = line.strip()
            if len(line) > 0 and not line.startswith("#"):  # Skip empty lines and comments
                deprecated_ids.add(line)
    except (OSError, IOError):
        return set()

    return deprecated_ids



from rich.progress import track

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.core import analysis_cache, root, versions_file, yaml
from panther_analysis_tool.gui.explore_gui import ExploreApp


def run() -> tuple[int, str]:
    root.chdir_to_project_root()
    analysis_cache.update_with_latest_panther_analysis(show_progress_bar=True)
    all_specs = load_panther_analysis_specs(show_progress_bar=True)
    user_spec_ids = load_user_specs(show_progress_bar=True)
    app = ExploreApp(
        all_specs=all_specs,
        user_spec_ids=user_spec_ids,
    )
    app.run()
    return 0, ""


def load_panther_analysis_specs(
    show_progress_bar: bool = False,
) -> list[analysis_utils.AnalysisItem]:
    yaml_loader = yaml.BlockStyleYAML()
    cache = analysis_cache.AnalysisCache()
    versions = versions_file.get_versions().versions

    specs: list[analysis_utils.AnalysisItem] = []
    for _id in track(
        cache.list_spec_ids(),
        description="Loading all specs:",
        disable=not show_progress_bar,
        transient=True,
    ):
        spec = cache.get_latest_spec(_id)
        if spec is None:
            continue
        if _id not in versions:
            continue

        yaml_content = yaml_loader.load(spec.spec)
        ver = versions[_id]

        item = analysis_utils.AnalysisItem(
            yaml_file_contents=yaml_content,
            raw_yaml_file_contents=spec.spec,
            yaml_file_path=ver.history[ver.version].yaml_file_path,
        )

        if "Filename" in yaml_content:
            python_file_contents = cache.get_file_for_spec(spec.id or -1, spec.version)
            item.python_file_contents = python_file_contents
            item.python_file_path = ver.history[ver.version].py_file_path

        specs.append(item)

    return specs


def load_user_specs(show_progress_bar: bool = False) -> set[str]:
    spec_ids: set[str] = set()
    for spec in track(
        analysis_utils.load_analysis_specs_ex(["."], [], True),
        description="Loading user analysis items:",
        disable=not show_progress_bar,
        transient=True,
    ):
        spec_ids.add(spec.analysis_id())
    return spec_ids



import ast
import base64
import datetime
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, List, Optional, TypeVar

import dateutil.parser

ResponseData = TypeVar("ResponseData")


class BackendError(Exception):
    permanent: bool = False


class PermanentBackendError(BackendError):
    permanent: bool = True


class UnsupportedEndpointError(Exception):
    pass


@dataclass(frozen=True)
class BackendResponse(Generic[ResponseData]):
    data: ResponseData
    status_code: int


@dataclass(frozen=True)
class BackendCheckResponse:
    success: bool
    message: str


@dataclass(frozen=True)
class BulkUploadParams:
    zip_bytes: bytes

    def encoded_bytes(self) -> str:
        return base64.b64encode(self.zip_bytes).decode("utf-8")


@dataclass(frozen=True)
class DeleteSavedQueriesParams:
    names: List[str]
    dry_run: bool
    include_detections: bool


@dataclass(frozen=True)
class DeleteDetectionsParams:
    ids: List[str]
    dry_run: bool
    include_saved_queries: bool


@dataclass(frozen=True)
class ListSchemasParams:
    is_managed: bool


@dataclass(frozen=True)
class UpdateSchemaParams:
    description: str
    name: str
    reference_url: str
    revision: int
    spec: str
    field_discovery_enabled: bool


@dataclass(frozen=True)
class BulkUploadStatistics:
    new: int
    total: int
    modified: int


@dataclass(frozen=True)
class BulkUploadIssue:
    path: str
    error_message: str

    @classmethod
    def from_json(cls, data: Optional[Dict[str, Any]]) -> Optional["BulkUploadIssue"]:
        if not data:
            return None

        return cls(path=data.get("path", ""), error_message=data.get("errorMessage", ""))


class BackendMultipartError(ABC):
    @abstractmethod
    def has_error(self) -> bool:
        pass

    @abstractmethod
    def get_error(self) -> Optional[str]:
        pass

    @abstractmethod
    def has_issues(self) -> bool:
        pass

    @abstractmethod
    def get_issues(self) -> List[BulkUploadIssue]:
        pass


@dataclass
class BulkUploadMultipartError(BackendMultipartError):
    error: str
    issues: List[BulkUploadIssue] = field(default_factory=lambda: [])

    @classmethod
    def from_jsons(cls, data: str) -> "BulkUploadMultipartError":
        try:
            return BulkUploadMultipartError.from_dict(json.loads(data))
        except json.decoder.JSONDecodeError:
            return BulkUploadMultipartError.from_dict({"error": data})

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]) -> "BulkUploadMultipartError":
        if not data:
            return cls(error="")

        raw_issues = data.get("issues", []) or []
        issues: List[BulkUploadIssue] = []
        for issue in raw_issues:
            issues.append(BulkUploadIssue.from_json(issue))  # type: ignore

        err = data.get("error") or ""
        err = parse_graphql_error(err)

        return cls(issues=issues, error=err)

    def has_error(self) -> bool:
        return self.error is not None and len(self.error) > 0

    def get_error(self) -> Optional[str]:
        return self.error

    def has_issues(self) -> bool:
        return self.issues is not None and len(self.issues) > 0

    def get_issues(self) -> List[BulkUploadIssue]:
        if not self.has_issues():
            return []

        return self.issues or []


@dataclass(frozen=True)
class BulkUploadValidateResult:
    issues: List[BulkUploadIssue] = field(default_factory=lambda: [])

    @classmethod
    def from_json(cls, data: Optional[Dict[str, Any]]) -> Optional["BulkUploadValidateResult"]:
        if not data:
            return None

        raw_issues = data.get("issues", []) or []
        issues: List[BulkUploadIssue] = []
        for issue in raw_issues:
            issues.append(BulkUploadIssue.from_json(issue))  # type: ignore

        return cls(issues=issues)


@dataclass(frozen=True)
class BulkUploadValidateStatusResponse(BackendMultipartError):
    status: str
    error: str
    result: Optional[BulkUploadValidateResult] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "BulkUploadValidateStatusResponse":
        result = BulkUploadValidateResult.from_json(data.get("result"))
        status = data.get("status") or ""
        err = data.get("error") or ""
        err = parse_graphql_error(err)
        return cls(result=result, status=status, error=err)

    def has_error(self) -> bool:
        return self.error is not None and len(self.error) > 0

    def get_error(self) -> Optional[str]:
        return self.error

    def has_issues(self) -> bool:
        return self.result is not None and len(self.result.issues) > 0

    def get_issues(self) -> List[BulkUploadIssue]:
        if not self.has_issues():
            return []

        return self.result.issues or []  # type: ignore

    def is_valid(self) -> bool:
        if self.has_error():
            return False

        if self.has_issues():
            return False

        return True


@dataclass(frozen=True)
class BulkUploadResponse:
    rules: BulkUploadStatistics
    queries: BulkUploadStatistics
    policies: BulkUploadStatistics
    data_models: BulkUploadStatistics
    lookup_tables: BulkUploadStatistics
    global_helpers: BulkUploadStatistics
    correlation_rules: BulkUploadStatistics


@dataclass(frozen=True)
class DeleteSavedQueriesResponse:
    names: List[str]
    detection_ids: List[str]


@dataclass(frozen=True)
class DeleteDetectionsResponse:
    ids: List[str]
    saved_query_names: List[str]


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class Schema:
    created_at: str
    description: str
    is_managed: bool
    name: str
    reference_url: str
    revision: int
    spec: str
    updated_at: str
    field_discovery_enabled: bool


@dataclass(frozen=True)
class ListSchemasResponse:
    schemas: List[Schema]


@dataclass(frozen=True)
class UpdateSchemaResponse:
    schema: Schema


@dataclass(frozen=True)
class TranspileToPythonParams:
    data: List[str]


@dataclass(frozen=True)
class TranspileToPythonResponse:
    transpiled_python: List[str]


@dataclass(frozen=True)
class GetRuleBodyParams:
    id: str  # pylint: disable=invalid-name


@dataclass(frozen=True)
class TestCorrelationRuleParams:
    yaml: str


@dataclass(frozen=True)
class TestCorrelationRuleResponse:
    results: List[Dict[str, Any]]


@dataclass(frozen=True)
class GetRuleBodyResponse:
    body: str
    tests: List[Dict[str, Any]]


@dataclass(frozen=True)
class TranspileFiltersParams:
    data: List[str]
    pat_version: str


@dataclass(frozen=True)
class TranspileFiltersResponse:
    transpiled_filters: List[str]


@dataclass(frozen=True)
class MetricsParams:
    from_date: datetime.datetime
    to_date: datetime.datetime
    interval_in_minutes: int


@dataclass(frozen=True)
class SeriesWithBreakdown:
    breakdown: Dict[str, Any]
    label: str
    value: float


@dataclass(frozen=True)
class MetricsResponse:
    bytes_processed_per_source: List[SeriesWithBreakdown]


@dataclass(frozen=True)
class PerfTestParams(BulkUploadParams):
    hour: datetime.datetime
    log_type: str
    timeout: datetime.datetime


@dataclass(frozen=True)
class TimeWindow:
    starts_at: datetime.datetime
    ends_at: datetime.datetime


@dataclass(frozen=True)
class SizeWindow:
    max_size_in_gb: int


@dataclass(frozen=True)
class DataWindow:
    size_window: Optional[SizeWindow]
    time_window: Optional[TimeWindow]


@dataclass(frozen=True)
class ReplayScope:
    log_types: List[str]
    data_window: DataWindow


@dataclass(frozen=True)
class ReplaySummary:
    total_alerts: int
    completed_at: Optional[datetime.datetime]
    rule_error_count: int
    rule_match_count: int
    evaluation_progress: int
    computation_progress: int
    log_data_size_estimate: int
    matches_processed_count: int
    events_processed_count: int
    events_matched_count: int
    read_time_nanos: int
    processing_time_nanos: int


@dataclass(frozen=True)
class ReplayResponse:
    replay_id: str
    state: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    completed_at: Optional[datetime.datetime]
    detection_id: str
    replay_type: str
    replay_scope: ReplayScope
    replay_summary: ReplaySummary

    @classmethod
    def from_json(cls, data: Dict[str, Any], replay_id: str, status: str) -> "ReplayResponse":
        scope = data.get("scope", {})
        data_window = scope.get("dataWindow", {})
        retrieved_size_window = data_window.get("size_window")
        size_window = (
            None
            if retrieved_size_window is None
            else SizeWindow(max_size_in_gb=retrieved_size_window.get("maxSizeInGB"))
        )
        retrieved_time_window = data_window.get("time_window")
        time_window = (
            None
            if retrieved_time_window is None
            else TimeWindow(
                starts_at=dateutil.parser.parse(retrieved_time_window.get("startsAt")),
                ends_at=dateutil.parser.parse(retrieved_time_window.get("endsAt")),
            )
        )
        summary = data.get("summary", {})

        return ReplayResponse(
            replay_id=replay_id,
            state=status,
            created_at=dateutil.parser.parse(data.get("createdAt")),  # type: ignore
            updated_at=dateutil.parser.parse(data.get("updatedAt")),  # type: ignore
            completed_at=parse_optional_time(data.get("completedAt")),
            detection_id=data.get("detectionId"),  # type: ignore
            replay_type=data.get("replayType"),  # type: ignore
            replay_scope=ReplayScope(
                log_types=scope.get("logTypes"),
                data_window=DataWindow(size_window=size_window, time_window=time_window),
            ),
            replay_summary=ReplaySummary(
                total_alerts=summary.get("totalAlerts"),
                completed_at=parse_optional_time(summary.get("completedAt")),
                rule_error_count=summary.get("ruleErrorCount"),
                rule_match_count=summary.get("ruleMatchCount"),
                evaluation_progress=summary.get("evaluationProgress"),
                computation_progress=summary.get("computationProgress"),
                log_data_size_estimate=summary.get("logDataSizeEstimate"),
                matches_processed_count=summary.get("matchesProcessedCount"),
                events_processed_count=summary.get("eventsProcessedCount"),
                events_matched_count=summary.get("eventsMatchedCount"),
                read_time_nanos=summary.get("readTimeNanos"),
                processing_time_nanos=summary.get("processingTimeNanos"),
            ),
        )


@dataclass(frozen=True)
class GenerateEnrichedEventParams:
    event: Dict[str, Any]  # json


@dataclass(frozen=True)
class GenerateEnrichedEventResponse:
    enriched_event: Dict[str, Any]  # json


@dataclass(frozen=True)
class FeatureFlagWithDefault:
    flag: str
    default_treatment: Optional[bool] = None


@dataclass(frozen=True)
class FeatureFlagTreatment:
    flag: str
    treatment: bool


@dataclass(frozen=True)
class FeatureFlagsParams:
    flags: List[FeatureFlagWithDefault]


@dataclass(frozen=True)
class FeatureFlagsResponse:
    flags: List[FeatureFlagTreatment]


class Client(ABC):
    @abstractmethod
    def check(self) -> BackendCheckResponse:
        pass

    @abstractmethod
    def async_bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        pass

    @abstractmethod
    def bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        pass

    @abstractmethod
    def bulk_validate(self, params: BulkUploadParams) -> BulkUploadValidateStatusResponse:
        pass

    @abstractmethod
    def get_rule_body(self, params: GetRuleBodyParams) -> BackendResponse[GetRuleBodyResponse]:
        pass

    @abstractmethod
    def test_correlation_rule(
        self, params: TestCorrelationRuleParams
    ) -> BackendResponse[TestCorrelationRuleResponse]:
        pass

    @abstractmethod
    def transpile_simple_detection_to_python(
        self, params: TranspileToPythonParams
    ) -> BackendResponse[TranspileToPythonResponse]:
        pass

    @abstractmethod
    def transpile_filters(
        self, params: TranspileFiltersParams
    ) -> BackendResponse[TranspileFiltersResponse]:
        pass

    @abstractmethod
    def delete_saved_queries(
        self, params: DeleteSavedQueriesParams
    ) -> BackendResponse[DeleteSavedQueriesResponse]:
        pass

    @abstractmethod
    def delete_detections(
        self, params: DeleteDetectionsParams
    ) -> BackendResponse[DeleteDetectionsResponse]:
        pass

    @abstractmethod
    def list_schemas(self, params: ListSchemasParams) -> BackendResponse[ListSchemasResponse]:
        pass

    @abstractmethod
    def update_schema(self, params: UpdateSchemaParams) -> BackendResponse[Any]:
        pass

    @abstractmethod
    def supports_async_uploads(self) -> bool:
        pass

    @abstractmethod
    def supports_bulk_validate(self) -> bool:
        pass

    @abstractmethod
    def supports_perf_test(self) -> bool:
        pass

    @abstractmethod
    def get_metrics(self, params: MetricsParams) -> BackendResponse[MetricsResponse]:
        pass

    @abstractmethod
    def run_perf_test(self, params: PerfTestParams) -> BackendResponse[ReplayResponse]:
        pass

    @abstractmethod
    def supports_enrich_test_data(self) -> bool:
        pass

    @abstractmethod
    def generate_enriched_event_input(
        self, params: GenerateEnrichedEventParams
    ) -> BackendResponse[GenerateEnrichedEventResponse]:
        pass

    @abstractmethod
    def feature_flags(self, params: FeatureFlagsParams) -> BackendResponse[FeatureFlagsResponse]:
        pass


def backend_response_failed(resp: BackendResponse) -> bool:
    return resp.status_code >= 400 or resp.data.get("statusCode", 0) >= 400


def to_bulk_upload_response(data: Any) -> BackendResponse[BulkUploadResponse]:
    default_stats = {"total": 0, "new": 0, "modified": 0}
    return BackendResponse(
        status_code=200,
        data=BulkUploadResponse(
            rules=BulkUploadStatistics(
                total=data.get("rules", default_stats).get("total"),
                new=data.get("rules", default_stats).get("new"),
                modified=data.get("rules", default_stats).get("modified"),
            ),
            queries=BulkUploadStatistics(
                total=data.get("queries", default_stats).get("total"),
                new=data.get("queries", default_stats).get("new"),
                modified=data.get("queries", default_stats).get("modified"),
            ),
            policies=BulkUploadStatistics(
                total=data.get("policies", default_stats).get("total"),
                new=data.get("policies", default_stats).get("new"),
                modified=data.get("policies", default_stats).get("modified"),
            ),
            data_models=BulkUploadStatistics(
                total=data.get("dataModels", default_stats).get("total"),
                new=data.get("dataModels", default_stats).get("new"),
                modified=data.get("dataModels", default_stats).get("modified"),
            ),
            lookup_tables=BulkUploadStatistics(
                total=data.get("lookupTables", default_stats).get("total"),
                new=data.get("lookupTables", default_stats).get("new"),
                modified=data.get("lookupTables", default_stats).get("modified"),
            ),
            global_helpers=BulkUploadStatistics(
                total=data.get("globalHelpers", default_stats).get("total"),
                new=data.get("globalHelpers", default_stats).get("new"),
                modified=data.get("globalHelpers", default_stats).get("modified"),
            ),
            correlation_rules=BulkUploadStatistics(
                total=data.get("correlationRules", default_stats).get("total"),
                new=data.get("correlationRules", default_stats).get("new"),
                modified=data.get("correlationRules", default_stats).get("modified"),
            ),
        ),
    )


def parse_optional_time(time: Optional[str]) -> Optional[datetime.datetime]:
    return None if time is None else dateutil.parser.parse(time)


def parse_graphql_error(err: str) -> str:
    """
    Attempt to take what might be a graphql error from the backend and turn it into a dict.
    If it is not a graphql error or could not be turned into a dict, the original error is
    returned.
    """
    try:
        # if the backend returns a graphqlerror, it has single quotes which json can't
        # handle. So we need to use literal eval here and then check if it has the
        # message field which an error would
        return ast.literal_eval(str(err)).get("message") or err
    except BaseException:  # pylint: disable=broad-except
        return err



import os
from dataclasses import dataclass
from fnmatch import fnmatch
from typing import Any, Dict, Generator, List, Optional, Set

from panther_analysis_tool.analysis_utils import (
    ClassifiedAnalysis,
    filter_analysis,
    load_analysis_specs,
    to_relative_path,
)
from panther_analysis_tool.constants import DATA_MODEL_LOCATION, HELPERS_LOCATION
from panther_analysis_tool.core.parse import Filter


@dataclass
class ZipChunk:
    """
    ZipChunk allows you to select which files should go into a given chunk
      - you can set patterns and types; this will match both cases
      - you can set patterns or types; this will match whichever value is set


    Patterns are Unix shell style:

        *       matches everything
        ?       matches any single character
        [seq]   matches any character in seq
        [!seq]  matches any char not in seq

        Note: if a pattern such as */folder is used ./folder/hey.txt will not match
        You will need a pattern that does */folder/* to match

    Types are the types we expect in AnalysisType of a given yaml
    """

    patterns: List[str]
    types: Set[str] = ()  # type: ignore
    max_size: Optional[int] = None


@dataclass
class ZipArgs:
    out: Any
    path: Any
    ignore_files: List[str]
    filters: List[Filter]
    filters_inverted: List[Filter]


def chunk_list(lst: List[Any], limit: int) -> Generator[List[Any], None, None]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), limit):
        yield lst[i : i + limit]


class ChunkFiles:
    chunk: ZipChunk
    files: List[str]
    primary_files: List[str]
    related_files: Dict[str, str]
    added_files: Dict[str, bool]

    def __init__(self, chunk: ZipChunk):
        self.chunk = chunk
        self.files = []
        self.primary_files = []
        self.related_files = {}
        self.added_files = {}

    def add_file(self, filename: str, parent: Optional[str] = None) -> None:
        if self.added_files.get(filename) is None:
            self.added_files[filename] = True
            self.files.append(filename)
            if parent is None:
                self.primary_files.append(filename)

        if parent:
            self.related_files[parent] = filename

    def can_chunk_further(self) -> bool:
        if self.chunk.max_size is None:
            return False
        return len(self.added_files) > self.chunk.max_size

    def matches_file(self, filename: str, spec: Optional[Dict[str, Any]] = None) -> bool:
        if len(self.chunk.types) > 0 and spec is not None and "AnalysisType" in spec:
            if spec["AnalysisType"] not in self.chunk.types:
                return False

            if len(self.chunk.patterns) == 0:
                return True

        for pattern in self.chunk.patterns:
            if fnmatch(filename, pattern):
                return True

        return False


def create_additional_chunks_if_needed(chunk_files: List[ChunkFiles]) -> List[ChunkFiles]:
    results: List[ChunkFiles] = []
    for chunked in chunk_files:
        if not chunked.can_chunk_further():
            results.append(chunked)
            continue
        for files in chunk_list(chunked.primary_files, chunked.chunk.max_size):  # type: ignore
            chunk = ChunkFiles(chunk=chunked.chunk)
            for file in files:
                chunk.add_file(file)
                related_file = chunked.related_files.get(file)
                if related_file is not None:
                    chunk.add_file(related_file)
            results.append(chunk)
    return results


def analysis_chunks(args: ZipArgs, chunks: Optional[List[ZipChunk]] = None) -> List[ChunkFiles]:
    """Generates all files that should be added to a zip file. If no chunks are provided
    a single chunk will be returned. Note: a file can be in multiple chunks if both chunks
    matches the file pattern
    :param args:
    :param chunks:
    :return:
    """
    analysis = analysis_for_chunks(args)
    return chunk_analysis(analysis, chunks)


def analysis_for_chunks(args: ZipArgs, no_helpers: bool = False) -> List[ClassifiedAnalysis]:
    analysis = []
    files: Set[str] = set()

    paths = [args.path]
    if not no_helpers:
        paths.extend([HELPERS_LOCATION, DATA_MODEL_LOCATION])
    for file_name, f_path, spec, _ in list(load_analysis_specs(paths, args.ignore_files)):
        if file_name not in files:
            analysis.append(ClassifiedAnalysis(file_name, f_path, spec))
            files.add(file_name)
            files.add("./" + file_name)

    return filter_analysis(analysis, args.filters, args.filters_inverted)


def chunk_analysis(
    analysis: List[ClassifiedAnalysis], chunks: Optional[List[ZipChunk]] = None
) -> List[ChunkFiles]:
    if chunks is None or len(chunks) == 0:
        chunks = [ZipChunk(patterns=["*"])]
    chunk_files = [ChunkFiles(f) for f in chunks]
    for item in analysis:
        analysis_spec_filename = item.file_name
        dir_name = item.dir_name
        analysis_spec = item.analysis_spec
        main_file = to_relative_path(analysis_spec_filename)
        for chunk in chunk_files:
            if chunk.matches_file(analysis_spec_filename, analysis_spec):
                chunk.add_file(main_file)
                if analysis_spec is not None and "Filename" in analysis_spec:
                    chunk.add_file(
                        to_relative_path(os.path.join(dir_name, analysis_spec["Filename"])),
                        parent=main_file,
                    )
    return create_additional_chunks_if_needed(chunk_files)



from typing import Optional

import pydantic
from ruamel import yaml

from panther_analysis_tool.constants import CACHED_VERSIONS_FILE_PATH


class AnalysisVersionHistoryItem(pydantic.BaseModel):
    """
    A model for an analysis version history item. The combined commit hash and file path
    can be used to get the analysis item from the panther-analysis repository.

    Attributes:
        commit_hash: The commit hash that the analysis item can be found in.
        yaml_file_path: The file path of the YAML file, relative to the root of the panther-analysis repository.
        py_file_path: The file path of the Python file, relative to the root of the panther-analysis repository.
    """

    commit_hash: str
    yaml_file_path: str
    py_file_path: Optional[str] = None


class AnalysisVersionItem(pydantic.BaseModel):
    """
    A model for an analysis version item in the versions file from panther-analysis.

    Attributes:
        history: A dictionary of version number to analysis version history item.
        sha256: The SHA256 hash of the analysis item, both YAML and Python files, that can be used to see if a file has changed.
        type: The type of the analysis item (e.g. rule, policy, datamodel, etc.).
        version: The version number of the analysis item.
    """

    history: dict[int, AnalysisVersionHistoryItem]
    sha256: str
    type: str
    version: int


class Versions(pydantic.BaseModel):
    """
    A model for the contents of the .versions.yml file.

    Attributes:
        versions: A dictionary of analysis ID to analysis version item.
    """

    versions: dict[str, AnalysisVersionItem]

    def has_item(self, analysis_id: str) -> bool:
        return analysis_id in self.versions

    def get_current_version_history_item(self, analysis_id: str) -> AnalysisVersionHistoryItem:
        return self.versions[analysis_id].history[self.versions[analysis_id].version]

    def get_current_version(self, analysis_id: str) -> int:
        if not self.has_item(analysis_id):
            raise ValueError(f"Analysis ID '{analysis_id}' not found in versions file")
        return self.versions[analysis_id].version


_VERSIONS: Optional[Versions] = None


def get_versions() -> Versions:
    global _VERSIONS
    if _VERSIONS is None:
        if not CACHED_VERSIONS_FILE_PATH.exists():
            raise FileNotFoundError(f"No versions file at {CACHED_VERSIONS_FILE_PATH}")

        with open(CACHED_VERSIONS_FILE_PATH, "rb") as version_file:
            versions = yaml.YAML(typ="safe").load(version_file)
            _VERSIONS = Versions(**versions)

    return _VERSIONS



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from copy import deepcopy
from typing import Any, Collection, Dict, Iterator, Type, Union, no_type_check


class ImmutableContainerMixin(ABC):
    """Base class for immutable collections"""

    _CONVERSIONS: Dict[Any, Any] = {}

    @classmethod
    @abstractmethod
    def mutable_type(cls) -> Any:
        """Specify the mutable type that corresponds to this immutable container class"""

    @classmethod
    @no_type_check
    def register(cls) -> None:
        """Register the corresponding mutable type for this class"""
        cls._CONVERSIONS[cls.mutable_type()] = cls

    def __init__(self, container: Any):
        self._container = self._shallow_copy(container)

    @abstractmethod
    def _shallow_copy(self, obj: Any) -> Any:
        """Creates a shallow copy of the given object"""

    def copy(self) -> Any:
        """
        Returns a deep copy of the wrapped container object, which is mutable.
        Useful for interoperability with dependencies that perform their own type-checking,
        or when mutation is necessary.
        NOTE: Deep-copying has a performance overhead, so use only when necessary.
        """
        return self.mutable_type()(deepcopy(self._container))

    @no_type_check
    def __getitem__(self, item):
        return self._ensure_immutable(self._container[item])

    def _ensure_immutable(self, value: Any) -> Any:
        immutable_type = self.__class__._CONVERSIONS.get(type(value))  # pylint: disable=W0212
        if immutable_type is not None:
            return immutable_type(value)
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._container})"

    def __len__(self) -> int:
        return self._container.__len__()

    def __iter__(self) -> Iterator:
        return iter(self._ensure_immutable(element) for element in self._container)


class ImmutableCaseInsensitiveDict(ImmutableContainerMixin, Mapping):  # pylint: disable=R0901
    """
    Read-only dictionary data type with case-insensitive lookup.
    Assumes keys are strings and can be converted to all-lowercase with .lower().
    """

    def __init__(self, container: Union[Mapping, Collection[str]]):
        super().__init__(container)
        self._case_insensitive_keymap: Dict[str, str] = {}
        self._keymap_key_iterator: Union[None, Iterator[str]] = None
        self._keymap_fully_built = False

    @classmethod
    def mutable_type(cls) -> Type[dict]:
        return dict

    def _shallow_copy(self, obj: dict) -> dict:
        return obj.copy()

    def to_dict(self) -> dict:
        """
        Create a deep copy as a mutable dictionary.
        """
        return self.copy()

    def __getitem__(self, item: str) -> Any:
        # Try with the given key first (speeds up case-sensitive matches)
        # If the wrapped container is empty, which is possible for schemaless fields,
        # reraise early.
        if item in self._container or not self._container:
            return super().__getitem__(item)

        original_key = self._case_insensitive_lookup(item)

        # If no entry exists, raise a KeyError containing the given name
        # in order to retain compatibility with `__contains__` and `get` methods
        if original_key is None:
            raise KeyError(item)

        # The key exists in the container, so call the parent method
        # with the original key, in order to ensure the
        # inherited immutability as well
        return super().__getitem__(original_key)

    def _case_insensitive_lookup(self, key: str) -> Union[None, str]:
        # Convert to lowercase and check if a case-insensitive entry exists in the current map.
        # Note that for relatively small keys (up to 15-20 characters),
        # there is no point in caching the lowercase conversion, as the time to fetch
        # from a dictionary is nearly the same.
        lowercase_key = key.lower()
        original_key = self._case_insensitive_keymap.get(lowercase_key)
        if original_key is not None:
            return original_key

        # If the key was not found we need to continue building the key map,
        # unless the build has been already completed
        if original_key is None and self._keymap_fully_built:
            return None

        return self._build_case_insensitive_keymap(lowercase_key)

    def _build_case_insensitive_keymap(self, lowercase_key: str) -> Union[None, str]:
        # Attribute access in a loop has some cost, which can be reduced
        # by assigning to a local variable
        keymap = self._case_insensitive_keymap
        if self._keymap_key_iterator is None:
            self._keymap_key_iterator = iter(self._container)

        matched_key = None
        for original_key in self._keymap_key_iterator:
            case_insensitive_key = original_key.lower()
            # In case of collisions retain only the first key mapping
            mapped_key = keymap.setdefault(case_insensitive_key, original_key)
            # Build the key map incrementally - speeds up lookups when
            # the first keys in order are used.
            if case_insensitive_key == lowercase_key and mapped_key == original_key:
                matched_key = original_key
                break

        if len(self._case_insensitive_keymap) == len(self._container):
            self._keymap_fully_built = True

        return matched_key


class ImmutableList(ImmutableContainerMixin, Sequence):  # pylint: disable=R0901
    """Read-only sequence data type"""

    @classmethod
    def mutable_type(cls) -> Type[list]:
        return list

    def _shallow_copy(self, obj: list) -> tuple:
        return tuple(obj)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self._container == other._container

        if isinstance(other, (tuple, list)):
            # Allow comparison with lists and tuples
            return len(self._container) == len(other) and self._container == tuple(other)

        return False

    def as_list(self) -> list:
        """
        Create a deep copy as a mutable list.
        """
        return self.copy()


ImmutableList.register()
ImmutableCaseInsensitiveDict.register()


def json_encoder(obj: Any) -> Any:
    """
    Custom encoder for immutable objects

    :param obj: the object for JSON serialization
    :return: a JSON-serializable object
    """
    if isinstance(obj, ImmutableContainerMixin):
        return obj._container  # pylint: disable=W0212
    raise TypeError(f"object of type {type(obj)} is not JSON serializable")



import json
import logging
import os
import typing
from dataclasses import dataclass
from typing import Any, Dict, Optional

import boto3

from .client import (
    BackendCheckResponse,
    BackendError,
    BackendResponse,
    BulkUploadParams,
    BulkUploadResponse,
    BulkUploadValidateStatusResponse,
    Client,
    DeleteDetectionsParams,
    DeleteDetectionsResponse,
    DeleteSavedQueriesParams,
    DeleteSavedQueriesResponse,
    FeatureFlagsParams,
    FeatureFlagsResponse,
    GenerateEnrichedEventParams,
    GenerateEnrichedEventResponse,
    GetRuleBodyParams,
    GetRuleBodyResponse,
    ListSchemasParams,
    ListSchemasResponse,
    MetricsParams,
    MetricsResponse,
    PerfTestParams,
    PermanentBackendError,
    ReplayResponse,
    Schema,
    TestCorrelationRuleParams,
    TestCorrelationRuleResponse,
    TranspileFiltersParams,
    TranspileFiltersResponse,
    TranspileToPythonParams,
    TranspileToPythonResponse,
    UpdateSchemaParams,
    UpdateSchemaResponse,
    backend_response_failed,
    to_bulk_upload_response,
)
from .errors import is_retryable_error

LAMBDA_CLIENT_NAME = "lambda"
AWS_PROFILE_ENV_KEY = "AWS_PROFILE"


def decode_body(res: BackendResponse) -> typing.Any:
    try:
        return json.loads(res.data["body"])
    except json.decoder.JSONDecodeError as decode_error:
        # pylint: disable=broad-exception-raised
        raise Exception(res.data["body"]) from decode_error


@dataclass(frozen=True)
class LambdaClientOpts:
    user_id: str
    aws_profile: Optional[str]
    datalake_lambda: str


class LambdaClient(Client):
    _user_id: str
    _lambda_client: boto3.client

    def __init__(self, opts: LambdaClientOpts):
        self._user_id = opts.user_id

        if opts.aws_profile is None:
            self._setup_client()
        else:
            self._setup_client_with_profile(opts.aws_profile)

    def _setup_client(self) -> None:
        self._lambda_client = boto3.client(LAMBDA_CLIENT_NAME)

    def _setup_client_with_profile(self, profile: str) -> None:
        logging.info("Using AWS profile: %s", profile)
        os.environ[AWS_PROFILE_ENV_KEY] = profile
        self._lambda_client = boto3.Session(profile_name=profile).client(LAMBDA_CLIENT_NAME)

    def check(self) -> BackendCheckResponse:
        return BackendCheckResponse(success=True, message="not implemented")

    def async_bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("async uploads not supported with lambda client")

    def bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        resp = self._parse_response(
            self._lambda_client.invoke(
                FunctionName="panther-analysis-api",
                InvocationType="RequestResponse",
                LogType="None",
                Payload=self._serialize_request(
                    {
                        "bulkUpload": {
                            "data": params.encoded_bytes(),
                            "userId": self._user_id,
                        },
                    }
                ),
            )
        )

        if backend_response_failed(resp):
            err = PermanentBackendError(resp.data)
            err.permanent = not is_retryable_error(resp.data)

            raise err

        body = decode_body(resp)
        return to_bulk_upload_response(body)

    def bulk_validate(self, params: BulkUploadParams) -> BulkUploadValidateStatusResponse:
        # pylint: disable=broad-exception-raised
        raise BaseException("bulk validate is not supported with lambda client")

    def delete_detections(
        self, params: DeleteDetectionsParams
    ) -> BackendResponse[DeleteDetectionsResponse]:
        entries = []
        for id_to_delete in params.ids:
            entries.append({"id": id_to_delete})

        res = self._parse_response(
            self._lambda_client.invoke(
                FunctionName="panther-analysis-api",
                InvocationType="RequestResponse",
                LogType="None",
                Payload=self._serialize_request(
                    {
                        "deleteDetections": {
                            "dryRun": params.dry_run,
                            "userId": self._user_id,
                            "entries": entries,
                            "includeSavedQueries": params.include_saved_queries,
                        }
                    }
                ),
            )
        )

        body = decode_body(res)

        return BackendResponse(
            status_code=res.status_code,
            data=DeleteDetectionsResponse(
                ids=body.get("ids") or [],
                saved_query_names=body.get("savedQueryNames") or [],
            ),
        )

    def delete_saved_queries(
        self, params: DeleteSavedQueriesParams
    ) -> BackendResponse[DeleteSavedQueriesResponse]:
        res = self._parse_response(
            self._lambda_client.invoke(
                FunctionName="panther-analysis-api",
                InvocationType="RequestResponse",
                LogType="None",
                Payload=self._serialize_request(
                    {
                        "deleteSavedQueriesByName": {
                            "names": params.names,
                            "dryRun": params.dry_run,
                            "userId": self._user_id,
                            "includeDetections": params.include_detections,
                        }
                    }
                ),
            )
        )

        body = decode_body(res)

        return BackendResponse(
            status_code=res.status_code,
            data=DeleteSavedQueriesResponse(
                names=body.get("names", []),
                detection_ids=body.get("detectionIds", []),
            ),
        )

    def list_schemas(self, params: ListSchemasParams) -> BackendResponse[ListSchemasResponse]:
        res = self._parse_response(
            self._lambda_client.invoke(
                FunctionName="panther-logtypes-api",
                InvocationType="RequestResponse",
                Payload=self._serialize_request(
                    {
                        "ListSchemas": {"isManaged": params.is_managed},
                    }
                ),
            )
        )
        if res.data.get("error") is not None:
            raise BackendError(res.data.get("error"))

        schemas = []
        for result in res.data["results"]:
            schemas.append(
                Schema(
                    created_at=result.get("createdAt", ""),
                    description=result.get("description", ""),
                    is_managed=result.get("isManaged", False),
                    name=result.get("name", ""),
                    reference_url=result.get("referenceURL", ""),
                    revision=result.get("revision", ""),
                    spec=result.get("spec", ""),
                    updated_at=result.get("updatedAt", ""),
                    field_discovery_enabled=result.get("fieldDiscoveryEnabled", False),
                )
            )

        return BackendResponse(status_code=200, data=ListSchemasResponse(schemas=schemas))

    def update_schema(self, params: UpdateSchemaParams) -> BackendResponse:
        res = self._parse_response(
            self._lambda_client.invoke(
                FunctionName="panther-logtypes-api",
                InvocationType="RequestResponse",
                Payload=self._serialize_request(
                    {
                        "PutUserSchema": {
                            "description": params.description,
                            "name": params.name,
                            "reference_url": params.reference_url,
                            "revision": params.revision,
                            "spec": params.spec,
                            "fieldDiscoveryEnabled": params.field_discovery_enabled,
                        }
                    }
                ),
            )
        )
        if res.data.get("error") is not None:
            raise BackendError(res.data.get("error"))

        schema = res.data.get("result", {})
        return BackendResponse(
            status_code=200,
            data=UpdateSchemaResponse(
                schema=Schema(
                    created_at=schema.get("createdAt", ""),
                    description=schema.get("description", ""),
                    is_managed=schema.get("isManaged", False),
                    name=schema.get("name", ""),
                    reference_url=schema.get("referenceURL", ""),
                    revision=schema.get("revision", ""),
                    spec=schema.get("spec", ""),
                    updated_at=schema.get("updatedAt", ""),
                    field_discovery_enabled=schema.get("fieldDiscoveryEnabled", False),
                )
            ),
        )

    def transpile_simple_detection_to_python(
        self, params: TranspileToPythonParams
    ) -> BackendResponse[TranspileToPythonResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException(
            "transpile simple detections to python is not supported with lambda client"
        )

    def test_correlation_rule(
        self, params: TestCorrelationRuleParams
    ) -> BackendResponse[TestCorrelationRuleResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("test correlation rule is not supported with lambda client")

    def transpile_filters(
        self, params: TranspileFiltersParams
    ) -> BackendResponse[TranspileFiltersResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("transpile filters is not supported with lambda client")

    def get_rule_body(self, params: GetRuleBodyParams) -> BackendResponse[GetRuleBodyResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("get rule body is not supported with lambda client")

    @staticmethod
    def _serialize_request(data: Dict[str, Any]) -> str:
        logging.debug(">>> %s", data)
        return json.dumps(data, allow_nan=False)

    @staticmethod
    def _parse_response(response: Dict[str, Any]) -> BackendResponse[Dict[str, Any]]:
        logging.debug("<<< %s", response)
        payload_str = response["Payload"].read().decode("utf-8")
        logging.debug("<<< %s", payload_str)

        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        payload = json.loads(payload_str)

        if status_code > 299 or response.get("FunctionError") == "Unhandled":
            logging.warning("backend error received: %s", payload)
            raise BackendError(payload.get("errorMessage", "unknown error"))

        return BackendResponse(
            data=payload,
            status_code=status_code,
        )

    def supports_async_uploads(self) -> bool:
        return False

    def supports_bulk_validate(self) -> bool:
        return False

    def supports_perf_test(self) -> bool:
        return False

    def get_metrics(self, params: MetricsParams) -> BackendResponse[MetricsResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("get metrics is not supported with lambda client")

    def run_perf_test(self, params: PerfTestParams) -> BackendResponse[ReplayResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("run perf test is not supported with lambda client")

    def supports_enrich_test_data(self) -> bool:
        return False

    def generate_enriched_event_input(
        self, params: GenerateEnrichedEventParams
    ) -> BackendResponse[GenerateEnrichedEventResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("enrich-test-data is not supported with lambda client")

    def feature_flags(self, params: FeatureFlagsParams) -> BackendResponse[FeatureFlagsResponse]:
        # pylint: disable=broad-exception-raised
        raise BaseException("feature-flags is not supported with lambda client")



import dataclasses
import logging
import pathlib
from typing import Tuple

from rich.progress import track

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.analysis_utils import load_analysis_specs_ex
from panther_analysis_tool.constants import AutoAcceptOption
from panther_analysis_tool.core import (
    analysis_cache,
    file_editor,
    merge_item,
    root,
    yaml,
)


@dataclasses.dataclass
class MergeArgs:
    analysis_id: str
    editor: str | None = None
    auto_accept: AutoAcceptOption | None = None
    write_merge_conflicts: bool = False


def run(args: MergeArgs) -> Tuple[int, str]:
    try:
        root.chdir_to_project_root()
        return merge_analysis(
            args.analysis_id, args.editor, args.auto_accept, args.write_merge_conflicts
        )
    except file_editor.EditorCommandNotFoundError as err:
        return 1, str(err)


def merge_analysis(
    analysis_id: str | None,
    editor: str | None,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
) -> Tuple[int, str]:
    # load all user analysis specs
    user_specs = list(load_analysis_specs_ex(["."], [], True))

    mergeable_items = get_mergeable_items(analysis_id, user_specs)
    if not mergeable_items and analysis_id is None:
        print("Nothing to merge.")
        return 0, ""

    if not mergeable_items and analysis_id is not None:
        spec_ids = [spec.analysis_id() for spec in user_specs]
        if analysis_id not in spec_ids:
            print(f"Analysis ID '{analysis_id}' not found in user analysis items.")
            return 0, ""
        print(f"Analysis ID '{analysis_id}' does not need merging.")
        return 0, ""

    merge_items(
        mergeable_items,
        analysis_id,
        editor,
        auto_accept,
        show_progress_bar=False,
        write_merge_conflicts=write_merge_conflicts,
    )

    return 0, ""


def get_mergeable_items(
    analysis_id: str | None,
    user_specs: list[analysis_utils.LoadAnalysisSpecsResult],
) -> list[merge_item.MergeableItem]:
    """
    Get all mergeable items. An analysis item is mergeable if it has a BaseVersion and the BaseVersion is less than the latest version
    in the analysis cache. If an analysis_id is provided, only the item that matches the analysis_id is returned.

    Args:
        analysis_id: The analysis id to filter by. If None, all mergeable items are returned.

    Returns:
        A list of MergeableItem objects that contain the user's item, the latest Panther item, and the base Panther item.
    """
    cache = analysis_cache.AnalysisCache()
    mergeable_items: list[merge_item.MergeableItem] = []

    if not user_specs:
        return mergeable_items

    for user_spec in user_specs:
        # New loader per spec so roundtrip state is not shared (see load_analysis_specs_ex).
        yaml_loader = yaml.BlockStyleYAML()
        user_spec_id = user_spec.analysis_id()
        if analysis_id is not None and analysis_id != user_spec_id:
            # user specified an analysis id, only merge that one
            continue

        # load the latest analysis item from the cache using the user spec's ID
        latest_spec = cache.get_latest_spec(user_spec_id)
        if latest_spec is None:
            # this happens with custom analysis items
            continue

        # if the user spec does not have a BaseVersion, add one
        if "BaseVersion" not in user_spec.analysis_spec:
            user_spec.analysis_spec["BaseVersion"] = 1

        # check if the user spec's BaseVersion is less than the latest version, skip merge if it is not
        user_spec_base_version: int = user_spec.analysis_spec.get("BaseVersion")
        if user_spec_base_version > latest_spec.version:
            logging.warning(
                "User spec %s has a base version greater than the latest version %s, skipping",
                user_spec_id,
                latest_spec.version,
            )
            continue
        if user_spec_base_version == latest_spec.version:
            continue

        # load the base analysis item from the cache using the user spec's BaseVersion
        base_spec = cache.get_spec_for_version(user_spec_id, user_spec_base_version)
        if base_spec is None:
            base_spec = analysis_cache.fetch_and_cache_old_version(
                cache, user_spec_id, user_spec_base_version
            )
            if base_spec is None:
                logging.warning(
                    "%s at version %s not found, skipping",
                    user_spec_id,
                    user_spec_base_version,
                )
                continue

        # load the python file for the user spec from the file system
        user_py: bytes | None = None
        py_path: pathlib.Path | None = None
        if user_spec.analysis_spec.get("Filename") is not None:
            py_path = user_spec.python_file_path()
            user_py = py_path.read_bytes() if py_path is not None else None

        mergeable_items.append(
            merge_item.MergeableItem(
                user_item=analysis_utils.AnalysisItem(
                    yaml_file_contents=user_spec.analysis_spec,
                    raw_yaml_file_contents=user_spec.raw_spec_file_content,
                    yaml_file_path=user_spec.spec_filename,
                    python_file_contents=user_py,
                    python_file_path=str(py_path) if py_path is not None else None,
                ),
                latest_panther_item=analysis_utils.AnalysisItem(
                    yaml_file_contents=yaml_loader.load(latest_spec.spec),
                    raw_yaml_file_contents=latest_spec.spec,
                    python_file_contents=cache.get_file_for_spec(
                        latest_spec.id or -1, latest_spec.version
                    ),
                ),
                base_panther_item=analysis_utils.AnalysisItem(
                    yaml_file_contents=yaml_loader.load(base_spec.spec),
                    raw_yaml_file_contents=base_spec.spec,
                    python_file_contents=cache.get_file_for_spec(
                        base_spec.id or -1, base_spec.version
                    ),
                ),
                latest_item_version=latest_spec.version,
            )
        )

    return mergeable_items


def merge_items(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    mergeable_items: list[merge_item.MergeableItem],
    analysis_id: str | None,
    editor: str | None,
    auto_accept: AutoAcceptOption | None = None,
    show_progress_bar: bool = False,
    write_merge_conflicts: bool = False,
    preview: bool = False,
) -> None:
    updated_item_ids: list[str] = []
    merge_conflict_item_ids: list[str] = []

    for mergeable_item in track(
        mergeable_items,
        description="Merging analysis items:",
        disable=not show_progress_bar,
        transient=True,
    ):
        # New loader per spec so roundtrip state is not shared (see load_analysis_specs_ex).
        yaml_loader = yaml.BlockStyleYAML()
        if analysis_id is not None and analysis_id != mergeable_item.user_item.analysis_id():
            # user specified an analysis id, only merge that one
            continue

        if preview:
            has_conflict = merge_item.merge_item_preview(mergeable_item)
            if has_conflict:
                merge_conflict_item_ids.append(mergeable_item.user_item.analysis_id())
            else:
                updated_item_ids.append(mergeable_item.user_item.analysis_id())
            continue

        has_conflict = merge_item.merge_item(
            mergeable_item, analysis_id is not None, editor, auto_accept, write_merge_conflicts
        )
        if has_conflict:
            merge_conflict_item_ids.append(mergeable_item.user_item.analysis_id())
            continue

        # once we know there are no conflicts, we can update the BaseVersion to say merging is complete
        yaml_file_path = pathlib.Path(mergeable_item.user_item.yaml_file_path or "")
        user_spec: dict = yaml_loader.load(yaml_file_path)
        user_spec["BaseVersion"] = mergeable_item.latest_item_version
        yaml_loader.dump(user_spec, yaml_file_path)

        # consider updated if no conflict with both files
        updated_item_ids.append(mergeable_item.user_item.analysis_id())

    if analysis_id is not None:
        return

    if preview:
        if len(updated_item_ids) > 0:
            print(
                f"Will update {len(updated_item_ids)} analysis item(s) to their latest Panther version:"
            )
            for item_id in updated_item_ids:
                print(f"  * {item_id}")
        if len(merge_conflict_item_ids) > 0:
            print(
                f"{len(merge_conflict_item_ids)} analysis item(s) will have merge conflicts when updated to latest Panther version:"
            )
            for conflict in merge_conflict_item_ids:
                print(f"  * {conflict}")
        return

    if len(updated_item_ids) > 0:
        print(f"Updated {len(updated_item_ids)} analysis item(s) with latest Panther version:")
        for item_id in updated_item_ids:
            print(f"  * {item_id}")
    if len(merge_conflict_item_ids) > 0:
        print(
            f"{len(merge_conflict_item_ids)} merge conflict(s) found, run `EDITOR=<editor> pat merge <id>` to resolve each conflict:"
        )
        for conflict in merge_conflict_item_ids:
            print(f"  * {conflict}")
    if len(updated_item_ids) > 0:
        print(
            "Run `git diff` to see the changes. Run `pat test` to test the changes and `pat upload` to upload them."
        )



import shlex
import threading
from collections import defaultdict
from typing import Any

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.widgets import DataTable, Footer, Input, Tree

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.command import install
from panther_analysis_tool.core import parse
from panther_analysis_tool.gui import widgets


class ExploreApp(App):
    """A Textual app to explore Panther Analysis content."""

    BINDINGS = [
        Binding("ctrl+d", "toggle_dark", "Toggle dark mode", show=True),
        Binding("ctrl+q", "quit", "Quit", show=True),
        Binding("tab", "switch_focus", "Switch focus", show=True, priority=True),
        Binding(
            "escape", "close_editors", "Back to analysis item explorer", show=True, priority=True
        ),
        Binding(
            "ctrl+v",  # ctrl+c displays a warning prompt about quitting
            "install_analysis_item",
            "Install and enable selected analysis item",
            show=True,
            priority=True,
        ),
        Binding("ctrl+c", "quit", "Quit"),
    ]

    CSS_PATH = "explore_gui.tcss"

    def __init__(
        self,
        all_specs: list[analysis_utils.AnalysisItem],
        user_spec_ids: set[str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.all_specs = all_specs
        self.view_editors = False
        self.selected_item: analysis_utils.AnalysisItem | None = None
        self.user_spec_ids = user_spec_ids

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Footer()
        yield Input(placeholder="Search by text or by filters...", id="search-input")

        yield Horizontal(
            Tree(id="tree", label="Analysis Content"),  # Will be populated in on_mount
            widgets.AnalysisItemDataTable(id="table", cursor_type="row", show_row_labels=True),
            id="main-content",
        )

        yield Horizontal(
            widgets.YAMLWindow(text="", id="yaml-window"),
            widgets.PythonWindow(text="", id="python-window"),
            id="code-windows",
        )

    def on_mount(self) -> None:
        """Populate the DataTable and Tree when the app starts."""
        self.all_specs.sort(key=lambda x: x.pretty_analysis_type() + x.analysis_id())

        table = self.query_one("#table", widgets.AnalysisItemDataTable)
        table.add_specs_to_table(
            [
                widgets.AnalysisDataTableItem(
                    item=item, user_has_item=item.analysis_id() in self.user_spec_ids
                )
                for item in self.all_specs
            ]
        )
        self.add_specs_to_tree(self.all_specs)
        self.query_one("#code-windows").styles.display = "none"

    def add_specs_to_tree(self, specs: list[analysis_utils.AnalysisItem]) -> None:
        """Add specs to the tree."""
        tree = self.query_one("#tree", Tree)
        tree.clear()
        tree.root.label = "Analysis Content"
        tree.root.expand()

        spec_by_type = defaultdict(list)

        for spec in specs:
            spec_by_type[spec.pretty_analysis_type(plural=True)].append(spec)

        for pretty_analysis_type, specs in spec_by_type.items():
            analysis_node = tree.root.add(pretty_analysis_type, expand=False)
            analysis_node.data = {"analysis_type": specs[0].analysis_type()}

            for spec in specs:
                analysis_node.add_leaf(spec.analysis_id())

    def analysis_item_by_id(self, id: str) -> analysis_utils.AnalysisItem:
        for spec in self.all_specs:
            if spec.analysis_id() == id:
                return spec
        raise ValueError(f"Analysis item with ID {id} not found in all_specs.")

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action should be enabled based on current state."""
        if action == "close_editors" or action == "install_analysis_item":
            # Only show the "Back to explorer" and "Install" binding when viewing editors
            return self.view_editors
        return True

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Handle row selection in the DataTable."""
        _, _, item_id, _ = event.data_table.get_row(event.row_key)
        item = self.analysis_item_by_id(item_id)
        self.view_editors = True
        self.switch_views(item)
        self.refresh_bindings()

    def on_input_changed(self, event: Input.Changed) -> None:
        """Handle search input changes to filter the table."""
        search_term = event.value.strip()
        try:
            search_terms = shlex.split(search_term)  # split with same rules as unix shell
        except ValueError:
            # ignore error and just don't search yet. Allows user to type a quote without closing it yet.
            return

        filters = parse.search_terms_to_filters(search_terms)
        table = self.query_one("#table", widgets.AnalysisItemDataTable)
        table.filter_by_filters(filters)

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        """Handle tree node selection to filter the table."""
        table = self.query_one("#table", widgets.AnalysisItemDataTable)
        # If it's a leaf node (individual item), show only that item
        if not event.node.children:
            selected_id = str(event.node.label)
            table.filter_by_id(selected_id)
        else:
            # If it's a parent node, show all items under it
            analysis_type = str((event.node.data or {})["analysis_type"])
            table.filter_by_type(analysis_type)

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.theme = "textual-dark" if self.theme == "textual-light" else "textual-light"

    def action_close_editors(self) -> None:
        """Close the editor view and return to the main content."""
        self.view_editors = False
        self.switch_views(None)
        self.refresh_bindings()

    def action_install_analysis_item(self) -> None:
        """Install and enable the selected analysis item."""
        if self.selected_item is None:
            self.notify("No analysis item selected.", severity="error")
            return

        self.notify(
            f"Installing {self.selected_item.pretty_analysis_type()} {self.selected_item.analysis_id()}..."
        )

        # Run the install operation in a background thread so the notification can display immediately
        threading.Thread(
            target=self._perform_install_in_thread,
            args=(self.selected_item.analysis_id(), self.selected_item.pretty_analysis_type()),
            daemon=True,
        ).start()

    def _mark_user_has_item_on_table(self, item_id: str) -> None:
        """Update the table to show the user has this item (must run on main thread)."""
        table = self.query_one("#table", widgets.AnalysisItemDataTable)
        table.mark_user_has_item(item_id)

    def _perform_install_in_thread(self, item_id: str, item_type: str) -> None:
        """Perform the actual install operation in a background thread."""
        try:
            install.install(item_id, [])
            # Use call_from_thread to safely update UI from the worker thread
            self.call_from_thread(self._mark_user_has_item_on_table, item_id)
            self.call_from_thread(
                self.notify,
                f"{item_type} {item_id} ready to use!",
            )
        except Exception as err:
            # Use call_from_thread to safely update UI from the worker thread
            self.call_from_thread(self.notify, str(err), severity="error")
        finally:
            self.call_from_thread(self._on_install_complete)

    def _on_install_complete(self) -> None:
        """Handle cleanup after install operation completes."""
        self.refresh_bindings()
        self.view_editors = False
        self.switch_views(None)

    def switch_views(self, item: analysis_utils.AnalysisItem | None) -> None:
        if self.view_editors and item is not None:
            self.selected_item = item
            yaml_window = self.query_one("#yaml-window", widgets.YAMLWindow)
            yaml_window.text = item.raw_yaml_file_contents.decode("utf-8")  # type: ignore
            yaml_window.focus()

            py_window = self.query_one("#python-window", widgets.PythonWindow)
            if item.python_file_contents is not None:
                py_window.text = item.python_file_contents.decode("utf-8")
                py_window.styles.display = "block"
            else:
                py_window.styles.display = "none"

            self.query_one("#code-windows").styles.display = "block"
            self.query_one("#main-content").styles.display = "none"
            self.query_one("#search-input").styles.display = "none"
        else:
            self.selected_item = None
            self.query_one("#code-windows").styles.display = "none"
            self.query_one("#main-content").styles.display = "block"
            self.query_one("#search-input").styles.display = "block"
            self.query_one("#table").focus()



from collections.abc import Mapping
from typing import Any, Optional

from panther_core.data_model import E_NO_DATA_MODEL_FOUND, DataModel
from panther_core.exceptions import PantherError

from .immutable import ImmutableCaseInsensitiveDict, json_encoder


class PantherEvent(ImmutableCaseInsensitiveDict):  # pylint: disable=R0901
    """Panther enriched event with unified data model (udm) access."""

    def __init__(self, event: Mapping, data_model: Optional[DataModel]):
        """Create data model lookups

        Args:
            event: Dictionary representing the event.
            data_model: the data model used for the LogType associated with this event
        """
        super().__init__(event)
        self.data_model = data_model

    def udm(self, key: str) -> Any:
        """Converts standard data model field to logtype field"""
        self._validate()
        # access values via standardized fields
        match = self._get_json_path(key)
        if match:
            return self._ensure_immutable(match.value)
        method = self._get_method(key)
        if method:
            return self._ensure_immutable(method(self._ensure_immutable(self._container)))
        # no matches, return None by default
        return None

    def udm_path(self, key: str) -> Optional[str]:
        """Returns the JSON path or method name for the mapped field"""
        self._validate()
        # access values via standardized fields
        match = self._get_json_path(key)
        if match:
            return str(match.full_path)
        method = self._get_method(key)
        if method:
            return getattr(method, "__name__", repr(method))
        # no matches, return None by default
        return None

    def _validate(self) -> None:
        if not self.data_model:
            raise PantherError(E_NO_DATA_MODEL_FOUND, self._container.get("p_log_type"))

    def _get_json_path(self, key: str) -> Any:
        if not self.data_model:  # makes linter happy, we never call this if not set
            return None
        if key not in self.data_model.paths:
            return None
        json_path = self.data_model.paths.get(key)
        if not json_path:
            return None
        matches = json_path.find(self._container)
        if len(matches) == 0:
            return None
        if len(matches) == 1:
            return matches[0]
        # pylint: disable=broad-exception-raised
        raise Exception(
            f"JSONPath [{json_path}] in DataModel [{self.data_model.data_model_id}], matched multiple fields."
        )

    def _get_method(self, key: str) -> Any:
        if not self.data_model:  # makes linter happy, we never call this if not set
            return None
        if key not in self.data_model.methods:
            return None
        method = self.data_model.methods.get(key)
        if callable(method):
            return method
        # no matches, return None by default
        return None

    json_encoder = json_encoder



import dataclasses
from typing import Dict, Tuple

from rich.progress import BarColumn, Progress, TextColumn, track

from panther_analysis_tool.analysis_utils import (
    LoadAnalysisSpecsResult,
    load_analysis_specs_ex,
)
from panther_analysis_tool.command import merge
from panther_analysis_tool.constants import AutoAcceptOption
from panther_analysis_tool.core import analysis_cache, install_item, root


@dataclasses.dataclass
class UpdateArgs:
    auto_accept: AutoAcceptOption | None = None
    write_merge_conflicts: bool = False
    preview: bool = False


def run(args: UpdateArgs) -> Tuple[int, str]:
    root.chdir_to_project_root()
    update(
        show_progress_bar=True,
        auto_accept=args.auto_accept,
        write_merge_conflicts=args.write_merge_conflicts,
        preview=args.preview,
    )
    return 0, ""


def update(
    show_progress_bar: bool = False,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
    preview: bool = False,
) -> None:
    # load specs
    user_analysis_specs: Dict[str, LoadAnalysisSpecsResult] = {}
    for spec in track(
        # this does not load anything from the .cache dir
        load_analysis_specs_ex(["."], [], True),
        description="Loading user analysis items:",
        disable=not show_progress_bar,
        transient=True,
    ):
        user_analysis_specs[spec.analysis_id()] = spec

    # populate cache
    analysis_cache.update_with_latest_panther_analysis(user_analysis_specs, show_progress_bar)

    # merge analysis items
    mergeable_items = merge.get_mergeable_items(None, list(user_analysis_specs.values()))
    if len(mergeable_items) > 0:
        merge.merge_items(
            mergeable_items=mergeable_items,
            analysis_id=None,
            editor=None,
            auto_accept=auto_accept,
            show_progress_bar=show_progress_bar,
            write_merge_conflicts=write_merge_conflicts,
            preview=preview,
        )

    # we need to check if the new merged python includes any
    # new global helper imports and install those so the new python works
    with Progress(
        TextColumn("Installing dependencies:"),
        BarColumn(),
        transient=True,
        disable=not show_progress_bar,
    ) as progress:
        task = progress.add_task("installing_dependencies", total=None)
        items = [item.merged_item for item in mergeable_items if item.merged_item is not None]
        install_item.install_deps(items)
        progress.update(task, completed=True)



import pathlib

from panther_analysis_tool import analysis_utils


def run() -> tuple[int, str]:
    fmt()
    return 0, ""


def fmt() -> None:
    for item in analysis_utils.load_analysis_specs_ex(["."], [], True):
        item.yaml_ctx.dump(item.analysis_spec, pathlib.Path(item.spec_filename))



import logging
import os
import re
from functools import reduce
from pathlib import Path
from typing import Any, Dict, List, Optional, TextIO, Union

import boto3
import requests
from packaging import version

from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.lambda_client import LambdaClient, LambdaClientOpts
from panther_analysis_tool.backend.public_api_client import (
    PublicAPIClient,
    PublicAPIClientOptions,
)
from panther_analysis_tool.constants import (
    PACKAGE_NAME,
    PANTHER_USER_ID,
    VERSION_STRING,
    AnalysisTypes,
)

UNKNOWN_VERSION = "unknown"


class BackendNotFoundException(Exception):
    pass


def get_latest_version() -> str:
    try:
        response = requests.get(f"https://pypi.org/pypi/{PACKAGE_NAME}/json", timeout=10)
        if response.status_code == 200:
            return response.json().get("info", {}).get("version", UNKNOWN_VERSION)
    except Exception:  # pylint: disable=broad-except
        logging.debug("Unable to determine latest version", exc_info=True)
    return UNKNOWN_VERSION


def is_latest(latest_version: str) -> bool:
    try:
        return version.parse(VERSION_STRING) >= version.parse(latest_version)
    except Exception:  # pylint: disable=broad-except
        logging.debug("Unable to determine latest version", exc_info=True)
    # if we run into any issues connecting or parsing the version,
    # we should just return True
    return True


def get_client(aws_profile: Optional[str], service: str) -> boto3.client:
    # optionally set env variable for profile passed as argument
    if aws_profile is not None:
        logging.info("Using AWS profile: %s", aws_profile)
        set_env("AWS_PROFILE", aws_profile)
        sess = boto3.Session(profile_name=aws_profile)
        client = sess.client(service)
    else:
        client = boto3.client(service)
    return client


def get_optional_backend(api_token: Optional[str], api_host: str) -> Optional[BackendClient]:
    if api_token:
        return PublicAPIClient(
            PublicAPIClientOptions(token=api_token, user_id=PANTHER_USER_ID, host=api_host)
        )

    return None


def get_api_backend(api_token: Optional[str], api_host: str) -> BackendClient:
    if not api_token:
        raise BackendNotFoundException("This function requires an API token. API token not found.")

    return PublicAPIClient(
        PublicAPIClientOptions(token=api_token, user_id=PANTHER_USER_ID, host=api_host)
    )


def get_backend(
    api_token: Optional[str], api_host: str, aws_profile: Optional[str]
) -> BackendClient:
    if api_token:
        return PublicAPIClient(
            PublicAPIClientOptions(token=api_token, user_id=PANTHER_USER_ID, host=api_host)
        )

    return LambdaClient(
        LambdaClientOpts(
            user_id=PANTHER_USER_ID,
            aws_profile=aws_profile,
            datalake_lambda="panther-snowflake-api",
        )
    )


def set_env(key: str, value: str) -> None:
    os.environ[key] = value


def convert_keys_to_lowercase(mapping: Dict[str, Any]) -> Dict[str, Any]:
    """A helper function for converting top-level dictionary keys to lowercase.
    Converting keys to lowercase maintains compatibility with how the backend
    behaves.

    Args:
        mapping: The dictionary.

    Returns:
        A new dictionary with each key converted with str.lower()
    """
    return {k.lower(): v for k, v in mapping.items()}


def deep_get(obj: Dict, path: List[str], default: Optional[Any] = None) -> Any:
    result = reduce(lambda val, key: val.get(key) if val else None, path, obj)  # type: ignore
    return result if result is not None else default


def convert_unicode(obj: Any) -> str:
    """Swap unicode 4 byte strings with arbitrary numbers of leading slashes with the actual character
    e.g. \\\\u003c => <"""
    string_to_convert = str(obj)
    return re.sub(r"\\*\\u([0-9a-f]{4})", lambda m: chr(int(m.group(1), 16)), string_to_convert)


def is_simple_detection(analysis_item: Dict[str, Any]) -> bool:
    return all(
        [
            analysis_item.get("Detection") is not None,
            is_correlation_rule(analysis_item) is False,
            is_policy(analysis_item) is False,
        ]
    )


def is_correlation_rule(analysis_item: Dict[str, Any]) -> bool:
    return analysis_item.get("AnalysisType") == AnalysisTypes.CORRELATION_RULE


def is_policy(analysis_item: Dict[str, Any]) -> bool:
    return analysis_item.get("AnalysisType") == AnalysisTypes.POLICY


def add_path_to_filename(output_path: str, filename: str) -> str:
    if output_path:
        if not os.path.isdir(output_path):
            logging.info(
                "Creating directory: %s",
                output_path,
            )
            os.makedirs(output_path)
        filename = f"{output_path.rstrip('/')}/{filename}"

    return filename


def log_and_write_to_file(msgs: List[str], filename: TextIO) -> None:
    for msg in msgs:
        filename.write(msg + "\n")
        logging.info(msg)


def get_spec_id(spec: dict) -> str:
    """Returns the ID of an analysis item."""
    # Two possible cases: item is an analysis item, or a schema
    if analysis_type := spec.get("AnalysisType", ""):
        id_keys = {
            "correlation_rule": "RuleID",
            "datamodel": "DataModelID",
            "global": "GlobalID",
            "lookup_table": "LookupName",
            "pack": "PackID",
            "policy": "PolicyID",
            "saved_query": "QueryName",
            "scheduled_query": "QueryName",
            "scheduled_rule": "RuleID",
            "rule": "RuleID",
        }
        return spec[id_keys[analysis_type]]
    if schema_name := spec.get("schema", ""):
        return schema_name
    # Not an analysis item:
    return ""


def get_imports(spec: dict, path: Union[str, Path, os.PathLike]) -> set[str]:
    """Checks if an analysis item has a Python file, and if so, what Python imports are used.
    'path' is the directory where this item resides."""
    # Define some useful regex expressions
    pattern_import = re.compile(r"(?:from (\w+) import \w+)|(?:import (\w+))")
    pattern_block_comment = re.compile(r'"{3}.*?"{3}', flags=re.DOTALL)
    pattern_line_comment = re.compile(r"#.*?\n")

    imports = set()  # Holds the set of imports
    if spec.get("Filename", "").endswith(".py"):
        pyfile = Path(path) / spec["Filename"]
        with pyfile.open("r") as file_:
            contents = file_.read()
            contents = pattern_block_comment.sub("", contents)
            contents = pattern_line_comment.sub("", contents)
            matches = pattern_import.finditer(contents)
            for match in matches:
                # Extract module name
                for module_name in match.groups():
                    if module_name is not None:
                        # Handle nested imports
                        module_name = module_name.split(".")[0]
                        imports.add(module_name)

    return imports


def get_recursive_mappings(id_: str, mapping: dict[str, set]) -> set:
    """Recursively fetch a set of all dictionary items by trying each referenced item as a key.
    Used in check_packs."""
    mappings = {id_}
    for submapping in mapping.get(id_, []):
        mappings.update(get_recursive_mappings(submapping, mapping))
    return mappings



import dataclasses
from typing import Any, Callable, DefaultDict, Dict, Generator, List, Optional

from panther_core.rule import Detection
from panther_core.testing import TestResult

from panther_analysis_tool.constants import AnalysisTypes
from panther_analysis_tool.util import is_simple_detection


# pylint: disable=too-many-return-statements
def analysis_id_field_name(analysis_type: str) -> str:
    """Returns the name of the field that holds the ID of this analysis item (e.g. RuleID, PolicyID)."""
    match analysis_type:
        case (
            AnalysisTypes.RULE
            | AnalysisTypes.SCHEDULED_RULE
            | AnalysisTypes.CORRELATION_RULE
            | AnalysisTypes.DERIVED
            | AnalysisTypes.SIMPLE_DETECTION
        ):
            return "RuleID"
        case AnalysisTypes.DATA_MODEL:
            return "DataModelID"
        case AnalysisTypes.POLICY:
            return "PolicyID"
        case AnalysisTypes.GLOBAL:
            return "GlobalID"
        case AnalysisTypes.SCHEDULED_QUERY | AnalysisTypes.SAVED_QUERY:
            return "QueryName"
        case AnalysisTypes.PACK:
            return "PackID"
        case AnalysisTypes.LOOKUP_TABLE:
            return "LookupName"
        case _:
            raise ValueError(f"Unsupported analysis type: {analysis_type}")


class ClassifiedAnalysis:
    def __init__(self, file_name: str, dir_name: str, analysis_spec: Dict[str, Any]):
        self.file_name = file_name
        self.dir_name = dir_name
        self.analysis_spec = analysis_spec

    def is_deprecated(self) -> bool:
        display_name = self.analysis_spec["DisplayName"]
        description = self.analysis_spec.get("Description", "")
        if "deprecated" in display_name.lower() or "deprecated" in description.lower():
            return True

        tags = {tag.lower() for tag in self.analysis_spec.get("Tags", [])}
        if "deprecated" in tags:
            return True
        return False

    def analysis_type(self) -> str:
        return self.analysis_spec["AnalysisType"] or ""

    def analysis_id(self) -> str:
        """Returns the analysis ID for a given analysis spec."""
        return self.analysis_spec.get(analysis_id_field_name(self.analysis_type()), "UNKNOWN_ID")


@dataclasses.dataclass
class ClassifiedAnalysisContainer:
    """Contains all classified analysis specs"""

    data_models: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    globals: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    detections: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    simple_detections: List[ClassifiedAnalysis] = dataclasses.field(
        init=False, default_factory=list
    )
    queries: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    lookup_tables: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    packs: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)

    def _self_as_list(self) -> List[List[ClassifiedAnalysis]]:
        return [
            self.data_models,
            self.globals,
            self.detections,
            self.simple_detections,
            self.queries,
            self.lookup_tables,
            self.packs,
        ]

    def empty(self) -> bool:
        return all(len(l) == 0 for l in self._self_as_list())

    def apply(
        self,
        func: Callable[[List[ClassifiedAnalysis]], List[ClassifiedAnalysis]],
    ) -> "ClassifiedAnalysisContainer":
        container = ClassifiedAnalysisContainer()
        container.data_models = func(self.data_models)
        container.globals = func(self.globals)
        container.detections = func(self.detections)
        container.simple_detections = func(self.simple_detections)
        container.queries = func(self.queries)
        container.lookup_tables = func(self.lookup_tables)
        container.packs = func(self.packs)
        return container

    def items(self) -> Generator[ClassifiedAnalysis, None, None]:
        for analysis_list in self._self_as_list():
            for classified in analysis_list:
                yield classified

    def add_classified_analysis(
        self, analysis_type: str, classified_analysis: ClassifiedAnalysis
    ) -> None:
        if is_simple_detection(classified_analysis.analysis_spec):
            self.simple_detections.append(classified_analysis)
        elif analysis_type in [
            AnalysisTypes.POLICY,
            AnalysisTypes.RULE,
            AnalysisTypes.SCHEDULED_RULE,
            AnalysisTypes.CORRELATION_RULE,
        ]:
            self.detections.append(classified_analysis)
        elif analysis_type == AnalysisTypes.DATA_MODEL:
            self.data_models.append(classified_analysis)
        elif analysis_type == AnalysisTypes.GLOBAL:
            self.globals.append(classified_analysis)
        elif analysis_type == AnalysisTypes.LOOKUP_TABLE:
            self.lookup_tables.append(classified_analysis)
        elif analysis_type == AnalysisTypes.PACK:
            self.packs.append(classified_analysis)
        elif analysis_type == AnalysisTypes.SAVED_QUERY:
            self.queries.append(classified_analysis)
        elif analysis_type == AnalysisTypes.SCHEDULED_QUERY:
            self.queries.append(classified_analysis)


@dataclasses.dataclass
class TestResultContainer:
    detection: Optional[Detection]
    result: TestResult
    failed_tests: DefaultDict[str, list]
    output: str


@dataclasses.dataclass
class TestResultsContainer:
    """A container for all test results"""

    passed: Dict[str, List[TestResultContainer]]
    errored: Dict[str, List[TestResultContainer]]



import os

from panther_analysis_tool.constants import PAT_ROOT_FILE_NAME
from panther_analysis_tool.core import git_helpers


def chdir_to_project_root() -> None:
    """
    Change the current working directory to the project root. This is the git root if `pat init`
    was run in the same directory as the git root, otherwise it is the directory where the `.pat-root` file is located.
    Assumes that project is a git repo and `.pat-root` is inside git repo if it has the file.
    """
    git_root = git_helpers.git_root()
    for root, _, files in os.walk(git_root):
        if PAT_ROOT_FILE_NAME in files:
            os.chdir(root)
            return

    os.chdir(git_root)



import atexit
import os
import shutil
import signal
import tempfile
from typing import Any
from uuid import uuid4


def setup_temp() -> None:
    """
    Creates a dedicated temporary directory for this process and cleans it up at exit
    """
    temp_dir = os.path.join(tempfile.gettempdir(), f"tmp-PAT-{uuid4()}")
    os.mkdir(temp_dir)
    tempfile.tempdir = temp_dir

    def clean_me_up(signum: int = 0, _frame: Any = None) -> None:
        shutil.rmtree(temp_dir, ignore_errors=True)

        # If this was called as a signal handler, re-raise the signal
        if signum in (signal.SIGINT, signal.SIGTERM):
            # reset to the default python handler and redeliver the signal so that the
            # python process exits properly
            signal.signal(signum, signal.SIG_DFL)
            os.kill(os.getpid(), signum)

    atexit.register(clean_me_up)
    signal.signal(signal.SIGINT, clean_me_up)
    signal.signal(signal.SIGTERM, clean_me_up)



import dataclasses
import pathlib
import shutil
import subprocess  # nosec:B404

from panther_analysis_tool.constants import DEFAULT_EDITOR


class EditorCommandNotFoundError(RuntimeError):
    """
    Raised when an editor command is not found in system path.
    """

    def __init__(self, editor: str) -> None:
        super().__init__(f"Editor command {editor} was not found in system path")


@dataclasses.dataclass
class MergeableFiles:
    users_file: pathlib.Path  # local
    base_file: pathlib.Path  # base
    panthers_file: pathlib.Path  # remote
    output_file: pathlib.Path  # output

    # used for editors that don't have a built in 3-way merge tool
    premerged_file: pathlib.Path

    def validate(self, premerged_required: bool) -> None:
        if premerged_required:
            if not self.premerged_file.exists():
                raise FileNotFoundError(f"Premerged file {self.premerged_file} not found")
        else:
            if not self.users_file.exists():
                raise FileNotFoundError(f"User's file {self.users_file} not found")
            if not self.base_file.exists():
                raise FileNotFoundError(f"Base file {self.base_file} not found")
            if not self.panthers_file.exists():
                raise FileNotFoundError(f"Panther's file {self.panthers_file} not found")
            if not self.output_file.exists():
                raise FileNotFoundError(f"Output file {self.output_file} not found")


def merge_files_in_editor(files: MergeableFiles, editor: str | None) -> bool:
    """
    Merge files in an editor.

    Args:
        files: The MergeableFiles object.

    Returns:
        True if the editor returns before the merge is solved because it is solved asynchronously, False otherwise.
    """
    editor = editor.lower() if editor else DEFAULT_EDITOR.lower()

    # test if command is within system path
    if shutil.which(editor) is None:
        raise EditorCommandNotFoundError(editor)

    args = [editor]
    async_edit = True
    match editor:
        # jetbrains editors, not all their products are included in this list
        case (
            "idea"
            | "pycharm"
            | "webstorm"
            | "phpstorm"
            | "rubymine"
            | "clion"
            | "goland"
            | "rider"
            | "appcode"
            | "rustrover"
            | "dataspell"
        ):
            files.validate(premerged_required=False)
            args.extend(
                [
                    "merge",
                    str(files.users_file),
                    str(files.panthers_file),
                    str(files.base_file),
                    str(files.output_file),
                ]
            )
        case "vi" | "vim":
            async_edit = False
            files.validate(premerged_required=True)
            files.output_file.write_bytes(files.premerged_file.read_bytes())
            args.append(str(files.output_file))
        case _:
            files.validate(premerged_required=True)
            files.output_file.write_bytes(files.premerged_file.read_bytes())
            args.append(str(files.output_file))

    subprocess.run(args, check=True)  # nosec:B603
    return async_edit



import logging
import pathlib
import subprocess  # nosec:B404
from typing import Tuple

from panther_analysis_tool.constants import PAT_ROOT_FILE_NAME
from panther_analysis_tool.core import analysis_cache, git_helpers


def run(working_dir: str) -> Tuple[int, str]:
    analysis_cache.update_with_latest_panther_analysis(show_progress_bar=True)
    setup_git_ignore()
    enable_rerere()
    pat_root_created = setup_pat_root(pathlib.Path(working_dir))
    print_ready_message(pat_root_created)
    return 0, ""


def setup_git_ignore() -> None:
    gitignore_file = git_helpers.git_root() / ".gitignore"
    if not gitignore_file.exists():
        print(".gitignore file created")
        gitignore_file.touch()

    ignorables = [
        {
            "name": "Panther settings",
            "values": [".panther_settings.*"],
        },
        {
            "name": "Python",
            "values": ["__pycache__/", "*.pyc", ".mypy_cache/", ".pytest_cache/"],
        },
        {
            "name": "Panther",
            "values": ["panther-analysis-*.zip", ".cache/"],
        },
        {
            "name": "IDEs",
            "values": [".vscode/", ".idea/"],
        },
    ]

    content = gitignore_file.read_text()

    # ensure ends with two newlines
    if content != "" and not content.endswith("\n\n"):
        if not content.endswith("\n"):
            content += "\n"
        content += "\n"

    for ignorable in ignorables:
        section_comment = f"# {ignorable['name']}\n"
        section_exists = section_comment in content
        missing_values = [v for v in ignorable["values"] if v not in content]

        if section_exists:
            if missing_values:
                insertion = "".join(f"{v}\n" for v in missing_values)
                content = content.replace(section_comment, section_comment + insertion, 1)
        else:
            content += section_comment
            for value in ignorable["values"]:
                if value not in content:
                    content += f"{value}\n"
            content += "\n"

    gitignore_file.write_text(content)


def enable_rerere() -> None:
    proc = subprocess.run(  # nosec:B603 B607
        ["git", "config", "rerere.enabled", "true"], check=True, capture_output=True
    )
    if proc.stderr is not None and proc.stderr.decode("utf-8") != "":
        logging.error("Failed to enable git rerere: %s", proc.stderr.decode("utf-8"))


def setup_pat_root(working_dir: pathlib.Path) -> bool:
    """
    `.pat-root` file is used to track the root of the Panther project if the root is not the same as the git root.
    Create a `.pat-root` file in the working directory if it doesn't already exist.
    If the working directory is the same as the git root, do not create the `.pat-root` file.

    Args:
        working_dir (pathlib.Path): The current working directory.

    Returns:
        bool: True if the `.pat-root` file was created, False otherwise.
    """
    git_root = pathlib.Path(git_helpers.git_root()).absolute()
    if git_root == working_dir.absolute():
        return False

    pat_root_file = working_dir / PAT_ROOT_FILE_NAME
    if pat_root_file.exists():
        return False

    pat_root_file.touch()
    pat_root_file.write_text(
        "# File created by Panther Analysis Tool to track the root of your Panther project. Please commit this file and do not delete it.\n"
    )
    return True


def print_ready_message(pat_root_created: bool) -> None:
    print("Project is ready to use!\n")

    if pat_root_created:
        print(
            "`init` command was not run in the same directory as the git root. "
            "`.pat-root` file created in current directory and will be used as the root "
            "of your Panther project for future `pat` commands.\n"
        )

    print("Next, you can start exploring and using Panther out of the box content:")
    print("  * Run `pat explore` to see the available content.")
    print("  * Run `pat install <id>` to install a detection you want to use in your repo.")
    print(
        "  * Run `pat install --filter LogTypes=<LOG_TYPE>` to install all detections for a given log type you have onboarded."
    )
    print(
        "  * Run `pat test` to test your content and then run `pat upload` to upload your content to Panther."
    )

    print(
        "\nOr run `pat --help` to see all available commands. "
        "Visit https://docs.panther.com/panther-developer-workflows/overview for more information."
    )



import dataclasses
from typing import Any

from panther_analysis_tool.constants import AutoAcceptOption


@dataclasses.dataclass
class DictMergeConflict:
    key: str
    cust_val: Any
    latest_val: Any
    base_val: Any


class Dict:
    """
    A class to merge two dictionaries with a 3-way merge.

    Attributes:
        customer_dict (dict): The customer dictionary.
        auto_accept (AutoAcceptOption | None): The auto accept option.
            If None (default), the user will be prompted to resolve the merge conflict.
            If AutoAcceptOption.YOURS, the customer value will be used.
            If AutoAcceptOption.PANTHERS, the latest value will be used.
    """

    def __init__(self, customer_dict: dict, auto_accept: AutoAcceptOption | None = None):
        self.customer_dict = customer_dict
        self.auto_accept = auto_accept

    def merge_dict(self, base_dict: dict, latest_dict: dict) -> list[DictMergeConflict]:
        """
        Merge the latest dict into the customer dict, using the base dict for a 3-way merge.

        Args:
            base_dict (dict): The base dictionary.
            latest_dict (dict): The latest dictionary.

        Returns:
            list[DictMergeConflict]: A list of DictMergeConflict objects. Each object contains the key, customer value, latest value,
            and base value for each key that has a merge conflict.
        """
        diff_keys = diff_dict_keys(self.customer_dict, latest_dict)

        diff_items: list[DictMergeConflict] = []

        for k, v in latest_dict.items():
            if k not in self.customer_dict:
                if k in base_dict:
                    # Key existed in base; customer removed it. Treat as conflict so the user
                    # can choose to keep it removed (yours) or re-add Panther's value (panthers).
                    if self.auto_accept == AutoAcceptOption.YOURS:
                        pass  # keep key absent
                    elif self.auto_accept == AutoAcceptOption.PANTHERS:
                        self.customer_dict[k] = v.strip() if isinstance(v, str) else v
                    else:
                        diff_items.append(
                            DictMergeConflict(
                                key=k, cust_val=None, latest_val=v, base_val=base_dict[k]
                            )
                        )
                else:
                    # New key from Panther; safe to add.
                    self.customer_dict[k] = v.strip() if isinstance(v, str) else v

        for key in diff_keys:
            cust_val = self.customer_dict[key] if key in self.customer_dict else None
            latest_val = latest_dict[key] if key in latest_dict else None
            base_val = base_dict[key] if key in base_dict else None

            if base_val == latest_val and base_val != cust_val:
                # customer value changed but latest did not, use customer value
                self.customer_dict[key] = cust_val
                continue
            elif base_val == cust_val and base_val != latest_val:
                # latest value changed but customer did not, use latest value
                self.customer_dict[key] = latest_val
                continue

            if self.auto_accept == AutoAcceptOption.YOURS:
                self.customer_dict[key] = cust_val
                continue
            elif self.auto_accept == AutoAcceptOption.PANTHERS:
                self.customer_dict[key] = latest_val
                continue

            # customer and latest values changed, add to diff items to be resolved by the user
            diff_items.append(DictMergeConflict(key, cust_val, latest_val, base_val))

        return diff_items


def diff_dict_keys(dict1: dict, dict2: dict) -> list[str]:
    """
    Diff the keys of two dictionaries. A key counts as different if the values are in both dicts and are different.

    Args:
        dict1: The first dictionary.
        dict2: The second dictionary.

    Returns:
        A list of keys that are different between the two dictionaries and are in both dicts.
    """
    diff = []
    for key in dict1:
        if key in dict2 and dict1[key] != dict2[key]:
            diff.append(key)
    return diff
